from onegov.core.orm.cache import orm_cached
from onegov.core.orm.session_manager import SessionManager
from sqlalchemy import event, inspect
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import object_session
from sqlalchemy_utils import TranslationHybrid
from zope.sqlalchemy import mark_changed


MISSING = object()


#: The base for all OneGov Core ORM Models
class ModelBase(object):

    #: set by :class:`onegov.core.orm.cache.OrmCacheDescriptor`, this attribute
    #: indicates if the current model was loaded from cache
    is_cached = False

    def __getstate__(self):
        """ Makes sure the session manager attached to the model is not
        pickled (stored in memcached). Not only does it not make sense to
        have connection state stored, it also leads to errors.

        Cached instances have to be merged back into the session as a result,
        but that's something you have to do anyway when pickling sqlalchemy
        models.

        """
        state = self.__dict__.copy()

        for key in ('session_manager', ):
            if key in state:
                del state[key]

        return state

    def __setstate__(self, state):
        self.__dict__ = state

    @classmethod
    def get_polymorphic_class(cls, identity_value, default=MISSING):
        """ Returns the polymorphic class if it exists, given the value
        of the polymorphic identity.

        Asserts that the identity is actually found, unless a default is
        provided.

        """
        mapper = inspect(cls).polymorphic_map.get(identity_value)

        if default is MISSING:
            assert mapper, "No such polymorphic_identity: {}".format(
                identity_value
            )

        return mapper and mapper.class_ or default


Base = declarative_base(cls=ModelBase)

#: A translation hybrid integrated with OneGov Core. See also:
#: http://sqlalchemy-utils.readthedocs.org/en/latest/internationalization.html
translation_hybrid = TranslationHybrid(
    current_locale=lambda obj: obj.session_manager.current_locale,
    default_locale=lambda obj: obj.session_manager.default_locale,
)


def find_models(base, is_match):
    """ Finds the ORM models in the given ORM base class that match a filter.

    The filter is called with each class in the instance and it is supposed
    to return True if it matches.

    For example, find all SQLAlchemy models that use
    :class:`~onegov.core.orm.mixins.ContentMixin`::

        from onegov.core.orm.mixins import ContentMixin
        find_models(base, is_match=lambda cls: issubclass(cls, ContentMixin))

    """
    for cls in base.__subclasses__():
        if is_match(cls):
            yield cls

        for cls in find_models(cls, is_match):
            yield cls


def configure_listener(cls, key, instance):
    """ The zope.sqlalchemy transaction mechanism doesn't recognize changes to
    cached objects. The following code intercepts all object changes and marks
    the transaction as changed if there was a change to a cached object.

    """

    def mark_as_changed(obj, *args, **kwargs):
        if obj.is_cached:
            mark_changed(object_session(obj))

    event.listen(instance, 'append', mark_as_changed)
    event.listen(instance, 'remove', mark_as_changed)
    event.listen(instance, 'set', mark_as_changed)
    event.listen(instance, 'init_collection', mark_as_changed)
    event.listen(instance, 'dispose_collection', mark_as_changed)


event.listen(ModelBase, 'attribute_instrument', configure_listener)


__all__ = [
    'Base', 'SessionManager', 'translation_hybrid', 'find_models', 'orm_cached'
]



from bleach.sanitizer import Cleaner


# html tags allowed by bleach
SANE_HTML_TAGS = [
    'a',
    'abbr',
    'b',
    'br',
    'blockquote',
    'code',
    'del',
    'div',
    'em',
    'i',
    'img',
    'hr',
    'li',
    'ol',
    'p',
    'pre',
    'strong',
    'sup',
    'span',
    'ul',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6'
]

# html attributes allowed by bleach
SANE_HTML_ATTRS = {
    'a': ['href', 'title'],
    'abbr': ['title', ],
    'acronym': ['title', ],
    'img': ['src', 'alt', 'title']
}


cleaner = Cleaner(
    tags=SANE_HTML_TAGS,
    attributes=SANE_HTML_ATTRS
)


def sanitize_html(html):
    """ Takes the given html and strips all but a whitelisted number of tags
    from it.

    """

    return cleaner.clean(html)


def sanitize_svg(svg):
    """ I couldn't find a good svg sanitiser function yet, so for now
    this function will be a no-op, though it will try to detect
    svg files which are harmful.

    I tried to go with bleach/html5lib, but the lack of xml namespace support
    makes those options a no go.

    In the future we want a proper SVG sanitiser here!

    """

    assert 'javascript:' not in svg
    assert 'CDATA' not in svg
    assert '<script>' not in svg
    assert 'Set-Cookie' not in svg

    return svg



from onegov.core.orm.types.hstore_type import HSTORE
from onegov.core.orm.types.json_type import JSON
from onegov.core.orm.types.utcdatetime_type import UTCDateTime
from onegov.core.orm.types.uuid_type import UUID
from onegov.core.orm.types.lowercase_text_type import LowercaseText

__all__ = ['HSTORE', 'JSON', 'UTCDateTime', 'UUID', 'LowercaseText']



import arrow
import babel.dates
import babel.numbers
import numbers
import sedate

from cached_property import cached_property
from datetime import datetime
from onegov.core import utils
from pytz import timezone


# I introduced a bug with arrow that should be fixed in a future release, this
# here is a workaround, until this is merged:
# https://github.com/crsmithdev/arrow/pull/410
from arrow.locales import SwissLocale
SwissLocale.timeframes['days'] = '{0} Tagen'


class Layout(object):
    """ Contains useful methods related to rendering pages in html. Think of it
    as an API that you can rely on in your templates.

    The idea is to provide basic layout functions here, if they are usful for
    any kind of html application. You should then extend the core layout
    classes with your own.

    """

    #: The timezone is currently fixed to 'Europe/Zurich' since all our
    #: business with onegov is done here. Once the need arises, we should
    #: lookup the timezone from the IP of the user, or use a javascript
    #: library that sets the timezone for the user session.
    #:
    #: There's also going to be the case where we want the timezone set
    #: specifically for a certain layout (say a reservation of a room, where
    #: the room's timezone is relevant). This is why this setting should
    #: remain close to the layout, and not necessarily close to the request.
    timezone = timezone('Europe/Zurich')

    #: Just like the timezone, these values are fixed for Switzerland now,
    #: though the non-numerical information is actually translated.
    #: Format:
    #
    #: http://www.unicode.org/reports/tr35/tr35-39/tr35-dates.html
    #: #Date_Format_Patterns
    #:
    #: Classes inheriting from :class:`Layout` may add their own formats, as
    #: long as they end in ``_format``. For example::
    #:
    #:     class MyLayout(Layout):
    #:         my_format = 'dd.MMMM'
    #:
    #:     MyLayout().format_date(dt, 'my')
    #:
    #: XXX this is not yet i18n and could be done better
    time_format = 'HH:mm'
    date_format = 'dd.MM.yyyy'
    datetime_format = 'dd.MM.yyyy HH:mm'

    date_long_format = 'dd. MMMM yyyy'
    datetime_long_format = 'd. MMMM yyyy HH:mm'
    weekday_long_format = 'EEEE'
    month_long_format = 'MMMM'

    def __init__(self, model, request):
        self.model = model
        self.request = request

    @cached_property
    def app(self):
        """ Returns the application behind the request. """
        return self.request.app

    def chunks(self, *args, **kwargs):
        """ See :func:`onegov.core.utils.chunks`. """

        return utils.chunks(*args, **kwargs)

    @cached_property
    def csrf_token(self):
        """ Returns a csrf token for use with DELETE links (forms do their
        own thing automatically).

        """
        token = self.request.new_csrf_token()
        return token.decode('utf-8') if isinstance(token, bytes) else token

    def csrf_protected_url(self, url):
        """ Adds a csrf token to the given url. """
        return utils.append_query_param(url, 'csrf-token', self.csrf_token)

    def format_date(self, dt, format):
        """ Takes a datetime and formats it according to local timezone and
        the given format.

        """
        if dt is None:
            return ''

        if getattr(dt, 'tzinfo', None) is not None:
            dt = self.timezone.normalize(dt.astimezone(self.timezone))

        if format == 'relative':
            dt = arrow.get(dt)

            try:
                return dt.humanize(locale=self.request.locale)
            except ValueError:
                return dt.humanize(locale=self.request.locale.split('_')[0])

        if hasattr(dt, 'hour'):
            formatter = babel.dates.format_datetime
        else:
            formatter = babel.dates.format_date

        return formatter(
            dt,
            format=getattr(self, format + '_format'),
            locale=self.request.locale
        )

    def isodate(self, date):
        """ Returns the given date in the ISO 8601 format. """
        return datetime.isoformat(date)

    def format_number(self, number, decimal_places=None):
        """ Takes the given numer and formats it according to locale.

        If the number is an integer, the default decimal places are 0,
        otherwise 2.

        """
        if decimal_places is None:
            if isinstance(number, numbers.Integral):
                decimal_places = 0
            else:
                decimal_places = 2

        locale = self.request.locale

        decimal = babel.numbers.get_decimal_symbol(locale)
        group = babel.numbers.get_group_symbol(locale)

        result = '{{:,.{}f}}'.format(decimal_places).format(number)
        return result.translate({ord(','): group, ord('.'): decimal})

    @property
    def view_name(self):
        """ Returns the view name of the current view, or None if it is the
        default view.

        Note: This relies on morepath internals and is experimental in nature!

        """
        return self.request.unconsumed and self.request.unconsumed[-1] or None

    def today(self):
        return self.now().date()

    def now(self):
        return sedate.to_timezone(sedate.utcnow(), self.timezone)


class ChameleonLayout(Layout):
    """ Extends the base layout class with methods related to chameleon
    template rendering.

    This class assumes the existance of two templates:

    - layout.pt -> Contains the page skeleton with headers, body and so on.
    - macros.pt -> Contains chameleon macros.

    """

    @cached_property
    def template_loader(self):
        """ Returns the chameleon template loader. """
        registry = self.app.config.template_engine_registry
        return registry._template_loaders['.pt']

    @cached_property
    def base(self):
        """ Returns the layout, which defines the base layout of all town
        pages. See ``templates/layout.pt``.

        """
        return self.template_loader['layout.pt']

    @cached_property
    def macros(self):
        """ Returns the macros, which offer often used html constructs.
        See ``templates/macros.pt``.

        """

        return self.template_loader.macros



from onegov.core.directives import query_form_class
from onegov.core.framework import Framework
from onegov.core.security import Secret
from webtest import TestApp as Client
from wtforms import Form


def test_form_directive():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.form(model=Root, form=Form)
    def handle_form(self, request, form):
        return ' '.join(('1', request.method, form.action))

    @App.form(model=Root, name='separate', request_method='GET', form=Form)
    def get_form(self, request, form):
        return ' '.join(('2', request.method, form.action))

    @App.form(model=Root, name='separate', request_method='POST', form=Form)
    def post_form(self, request, form):
        return ' '.join(('3', request.method, form.action))

    @App.form(model=Root, name='1', form=Form, permission=Secret)
    def handle_blocked_one(self, request, form):
        pass

    @App.form(model=Root, name='2', form=Form, permission=Secret,
              request_method='GET')
    def handle_blocked_two(self, request, form):
        pass

    @App.form(model=Root, name='3', form=Form, permission=Secret,
              request_method='POST')
    def handle_blocked_three(self, request, form):
        pass

    app = App()
    app.application_id = 'test'
    app.configure_application(
        identity_secure=False,
        disable_memcached=True
    )

    client = Client(app)
    assert client.get('/').text == '1 GET http://localhost/'
    assert client.post('/').text == '1 POST http://localhost/'
    assert client.get('/separate').text == '2 GET http://localhost/separate'
    assert client.post('/separate').text == '3 POST http://localhost/separate'

    assert client.get('/1', expect_errors=True).status_code == 403
    assert client.post('/1', expect_errors=True).status_code == 403
    assert client.get('/2', expect_errors=True).status_code == 403
    assert client.post('/2', expect_errors=True).status_code == 405
    assert client.get('/3', expect_errors=True).status_code == 405
    assert client.post('/3', expect_errors=True).status_code == 403


def test_query_form_class():

    class FormA(Form):
        pass

    class FormB(Form):
        pass

    def get_form_b(self, request):
        return FormB

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.form(model=Root, form=FormA)
    def handle_form(self, request, form):
        return ''

    @App.form(model=Root, name='b', form=get_form_b)
    def get_form(self, request, form):
        return ''

    @App.view(model=Root, name='assert-form-a')
    def assert_form_a(self, request):
        assert query_form_class(request, Root) is FormA

    @App.view(model=Root, name='assert-form-b')
    def assert_form_b(self, request):
        assert query_form_class(request, Root, name='b') is FormB

    @App.view(model=Root, name='assert-form-missing')
    def assert_form_c(self, request):
        assert query_form_class(request, Root, name='foobar') is None

    App.commit()

    app = App()
    app.application_id = 'test'
    app.configure_application(
        identity_secure=False,
        disable_memcached=True
    )

    client = Client(app)
    assert client.get('/assert-form-a')
    assert client.get('/assert-form-b')
    assert client.get('/assert-form-missing')



import math

from cached_property import cached_property
from sqlalchemy.inspection import inspect


class GenericCollection(object):

    def __init__(self, session):
        self.session = session

    @property
    def model_class(self):
        raise NotImplementedError

    @cached_property
    def primary_key(self):
        return inspect(self.model_class).primary_key[0]

    def query(self):
        return self.session.query(self.model_class)

    def by_id(self, id):
        return self.query().filter(self.primary_key == id).first()

    def add(self, **kwargs):
        item = self.model_class(**kwargs)

        self.session.add(item)
        self.session.flush()

        return item

    def delete(self, item):
        self.session.delete(item)
        self.session.flush()


class Pagination(object):
    """ Provides collections with pagination, if they implement a few
    documented properties and methods.

    See :class:`onegov.ticket.TicketCollection` for an example.

    """

    batch_size = 10

    def __eq__(self, other):
        """ Returns True if the current and the other Pagination instance
        are equal. Used to find the current page in a list of pages.

        """
        raise NotImplementedError

    def subset(self):
        """ Returns an SQLAlchemy query containing all records that should
        be considered for pagination.

        """
        raise NotImplementedError

    @cached_property
    def cached_subset(self):
        return self.subset()

    @property
    def page_index(self):
        """ Returns the current page index (starting at 0). """
        raise NotImplementedError

    def page_by_index(self, index):
        """ Returns the page at the given index. A page here means an instance
        of the class inheriting from the ``Pagination`` base class.

        """
        raise NotImplementedError

    def transform_batch_query(self, query):
        """ Allows subclasses to transform the given query before it is
        used to retrieve the batch. This is a good place to add additional
        loading that should only apply to the batch (say joining other
        values to the batch which are then not loaded by the whole query).

        """
        return query

    @cached_property
    def subset_count(self):
        """ Returns the total number of elements this pagination represents.

        """

        # the ordering is entirely unnecessary for a count, so remove it
        # to count things faster
        return self.cached_subset.order_by(None).count()

    @cached_property
    def batch(self):
        """ Returns the elements on the current page. """
        query = self.cached_subset.slice(
            self.offset, self.offset + self.batch_size
        )
        return self.transform_batch_query(query).all()

    @property
    def offset(self):
        """ Returns the offset applied to the current subset. """
        return self.page * self.batch_size

    @property
    def pages_count(self):
        """ Returns the number of pages. """
        if not self.batch_size:
            return 1
        return int(math.ceil(self.subset_count / self.batch_size))

    @property
    def pages(self):
        """ Yields all page objects of this Pagination. """
        for page in range(0, self.pages_count):
            yield self.page_by_index(page)

    @property
    def previous(self):
        """ Returns the previous page or None. """
        if self.page - 1 >= 0:
            return self.page_by_index(self.page - 1)

    @property
    def next(self):
        """ Returns the next page or None. """
        if self.page + 1 < self.pages_count:
            return self.page_by_index(self.page + 1)



from onegov.core.orm.mixins.content import (
    content_property,
    ContentMixin,
    data_property,
    meta_property,
)
from onegov.core.orm.mixins.timestamp import TimestampMixin

__all__ = [
    'content_property',
    'ContentMixin',
    'data_property',
    'meta_property',
    'TimestampMixin'
]



# We use our own custom json implementation. In the libres library we made this
# configurable. Since onegov.core is a framework we don't do that though, we
# want all onegov.core applications with the same framework version to be able
# to read each others json.
#
# Therefore we use a common denominator kind of json encoder/decoder.
from onegov.core import custom_json as json

from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.types import TypeDecorator, TEXT


class JSON(TypeDecorator):
    """ A JSON type for postgres that's backwards compatible with Postgres 9.1.

    In the future this will be replaced by JSON (or JSONB) though that requires
    that we require a later Postgres release. It would also mean that all
    json types have to be migrated.

    This type is associated with Sqlalchemy's MutableDict, so changes made to
    the dict are reflected in the database.

    See http://docs.sqlalchemy.org/en/rel_0_9/orm/extensions/mutable.html

    """

    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)

        return value


MutableDict.associate_with(JSON)



import json
import onegov.core
import os.path
import pytest
import transaction

from onegov.core import utils
from onegov.core.orm import SessionManager
from onegov.core.orm.types import HSTORE
from sqlalchemy import Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from unittest.mock import patch
from uuid import uuid4


def test_normalize_for_url():
    assert utils.normalize_for_url('asdf') == 'asdf'
    assert utils.normalize_for_url('Asdf') == 'asdf'
    assert utils.normalize_for_url('A S d f') == 'a-s-d-f'
    assert utils.normalize_for_url('far  away') == 'far-away'
    assert utils.normalize_for_url('währung') == 'waehrung'
    assert utils.normalize_for_url('grün') == 'gruen'
    assert utils.normalize_for_url('rötlich') == 'roetlich'
    assert utils.normalize_for_url('one/two') == 'one-two'
    assert utils.normalize_for_url('far / away') == 'far-away'
    assert utils.normalize_for_url('far <away>') == 'far-away'
    assert utils.normalize_for_url('far (away)') == 'far-away'
    assert utils.normalize_for_url('--ok--') == 'ok'
    assert utils.normalize_for_url('a...b..c.d') == 'a-b-c-d'


def test_lchop():
    assert utils.lchop('foobar', 'foo') == 'bar'
    assert utils.lchop('foobar', 'bar') == 'foobar'


def test_rchop():
    assert utils.rchop('foobar', 'foo') == 'foobar'
    assert utils.rchop('foobar', 'bar') == 'foo'
    assert utils.rchop('https://www.example.org/ex/amp/le', '/ex/amp/le') \
        == 'https://www.example.org'


def test_touch(temporary_directory):
    path = os.path.join(temporary_directory, 'test.txt')

    assert not os.path.isfile(path)

    utils.touch(path)

    assert os.path.isfile(path)

    with open(path, 'w') as f:
        f.write('asdf')

    utils.touch(path)

    with open(path, 'r') as f:
        assert f.read() == 'asdf'


def test_module_path():
    path = utils.module_path('onegov.core', 'utils.py')
    assert path == utils.module_path(onegov.core, 'utils.py')
    assert path == utils.module_path(onegov.core, '/utils.py')
    assert os.path.isfile(path)

    with pytest.raises(AssertionError):
        utils.module_path(onegov.core, '../passwd')


def test_linkify():
    # this is really bleach's job, but we want to run the codepath anyway
    assert utils.linkify('info@example.org')\
        == '<a href="mailto:info@example.org">info@example.org</a>'
    assert utils.linkify('https://google.ch')\
        == '<a href="https://google.ch" rel="nofollow">https://google.ch</a>'

    # by default, linkify sanitizes the text before linkifying it
    assert utils.linkify('info@example.org<br>')\
        == '<a href="mailto:info@example.org">info@example.org</a>&lt;br&gt;'

    # we can disable that however
    assert utils.linkify('info@example.org<br>', escape=False)\
        == '<a href="mailto:info@example.org">info@example.org</a><br>'


def test_increment_name():
    assert utils.increment_name('test') == 'test-1'
    assert utils.increment_name('test-2') == 'test-3'
    assert utils.increment_name('test2') == 'test2-1'
    assert utils.increment_name('test-1-1') == 'test-1-2'


def test_ensure_scheme():
    assert utils.ensure_scheme(None) is None
    assert utils.ensure_scheme('seantis.ch') == 'http://seantis.ch'
    assert utils.ensure_scheme('seantis.ch', 'https') == 'https://seantis.ch'

    assert utils.ensure_scheme('google.ch?q=onegov.cloud')\
        == 'http://google.ch?q=onegov.cloud'

    assert utils.ensure_scheme('https://abc.xyz') == 'https://abc.xyz'


def test_is_uuid():
    assert not utils.is_uuid(None)
    assert not utils.is_uuid('')
    assert not utils.is_uuid('asdf')
    assert utils.is_uuid(uuid4())
    assert utils.is_uuid(str(uuid4()))
    assert utils.is_uuid(uuid4().hex)


def test_is_non_string_iterable():
    assert utils.is_non_string_iterable([])
    assert utils.is_non_string_iterable(tuple())
    assert utils.is_non_string_iterable({})
    assert not utils.is_non_string_iterable('abc')
    assert not utils.is_non_string_iterable(b'abc')
    assert not utils.is_non_string_iterable(None)


def test_relative_url():
    assert utils.relative_url('https://www.google.ch/test') == '/test'
    assert utils.relative_url('https://usr:pwd@localhost:443/test') == '/test'
    assert utils.relative_url('/test') == '/test'
    assert utils.relative_url('/test?x=1&y=2') == '/test?x=1&y=2'
    assert utils.relative_url('/test?x=1&y=2#link') == '/test?x=1&y=2#link'


def test_is_subpath():
    assert utils.is_subpath('/', '/test')
    assert utils.is_subpath('/asdf', '/asdf/asdf')
    assert not utils.is_subpath('/asdf/', '/asdf')
    assert not utils.is_subpath('/a', '/b')
    assert not utils.is_subpath('/a', '/a/../b')


def test_is_sorted():
    assert utils.is_sorted('abc')
    assert not utils.is_sorted('aBc')
    assert utils.is_sorted('aBc', key=lambda i: i.lower())
    assert not utils.is_sorted('321')
    assert utils.is_sorted('321', reverse=True)


def test_get_unique_hstore_keys(postgres_dsn):

    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'

        id = Column(Integer, primary_key=True)
        _tags = Column(HSTORE, nullable=True)

        @property
        def tags(self):
            return set(self._tags.keys()) if self._tags else set()

        @tags.setter
        def tags(self, value):
            self._tags = {k: '' for k in value} if value else None

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('foo')

    assert utils.get_unique_hstore_keys(mgr.session(), Document._tags) == set()

    mgr.session().add(Document(tags=None))
    mgr.session().add(Document(tags=['foo', 'bar']))
    mgr.session().add(Document(tags=['foo', 'baz']))

    transaction.commit()

    assert utils.get_unique_hstore_keys(mgr.session(), Document._tags) == {
        'foo', 'bar', 'baz'
    }


def test_remove_repeated_spaces():

    assert utils.remove_repeated_spaces('  ') == ' '
    assert utils.remove_repeated_spaces('a b') == 'a b'
    assert utils.remove_repeated_spaces('a       b') == 'a b'
    assert utils.remove_repeated_spaces((' x  ')) == ' x '


def test_post_thread(session):
    with patch('urllib.request.urlopen') as urlopen:
        url = 'https://example.com/post'
        data = {'key': 'ä$j', 'b': 2}
        data = json.dumps(data).encode('utf-8')
        headers = (
            ('Content-type', 'application/json; charset=utf-8'),
            ('Content-length', len(data))
        )

        thread = utils.PostThread(url, data, headers)
        thread.start()
        thread.join()

        assert urlopen.called
        assert urlopen.call_args[0][0].get_full_url() == url
        assert urlopen.call_args[0][1] == data
        assert urlopen.call_args[0][0].headers == dict(headers)


def test_binary_dictionary():

    d = utils.binary_to_dictionary(b'foobar')
    assert d['filename'] is None
    assert d['mimetype'] == 'text/plain'
    assert d['size'] == 6

    d = utils.binary_to_dictionary(b'foobar', 'readme.txt')
    assert d['filename'] == 'readme.txt'
    assert d['mimetype'] == 'text/plain'
    assert d['size'] == 6

    assert utils.dictionary_to_binary(d) == b'foobar'


def test_safe_format():
    fmt = utils.safe_format

    assert fmt('hello [user]', {'user': 'admin'}) == 'hello admin'
    assert fmt('[ix]: [ix]', {'ix': 1}) == '1: 1'
    assert fmt('[[user]]', {'user': 'admin'}) == '[user]'
    assert fmt('[[[user]]]', {'user': 'admin'}) == '[admin]'
    assert fmt('[asdf]', {}) == ''
    assert fmt('[foo]', {'FOO': 'bar'}, adapt=str.upper) == 'bar'

    with pytest.raises(RuntimeError) as e:
        fmt('[foo[bar]]', {'foo[bar]': 'baz'})

    assert 'bracket inside bracket' in str(e)

    with pytest.raises(RuntimeError) as e:
        fmt('[secret]', {'secret': object()})

    assert 'type' in str(e)

    with pytest.raises(RuntimeError) as e:
        fmt('[asdf', {})

    assert 'Uneven' in str(e)

    with pytest.raises(RuntimeError) as e:
        fmt('[foo]', {}, raise_on_missing=True)

    assert 'is unknown' in str(e)



""" Provides tools and methods for internationalization (i18n).

Applications wishing to use i18n need to define two settings:

:i18n.localedirs:
    A list of gettext locale directories. The first directory is considered
    to be the main directory, all other directories are added to the
    translation object as fallbacks.

:i18n.default_locale:
    The fallback locale that is used if no locale more suitable to the user
    could be found.

For example::

    from onegov.core.framework import Framework
    from onegov.core import utils

    class App(Framework):
        pass

    @TownApp.setting(section='i18n', name='localedirs')
    def get_i18n_localedirs():
        return [
            utils.module_path('onegov.town', 'locale')
            utils.module_path('onegov.form', 'locale')
        ]

    @TownApp.setting(section='i18n', name='default_locale')
    def get_i18n_default_locale():
        return 'en'

"""

import gettext
import glob
import morepath
import os
import os.path
import polib
import re

from onegov.core.cache import lru_cache
from io import BytesIO
from onegov.core.framework import Framework, log
from onegov.core.utils import pairwise
from translationstring import ChameleonTranslate
from translationstring import Translator


# the language codes must be written thusly: de_CH, en_GB, en, fr and so on
# this is important because other libs like wtforms use the same scheme and
# will fail to deal with our langauges correctly if they differ in case
VALID_LANGUAGE_EXPRESSION = re.compile(r'^[a-z]{2}(_[A-Z]{2})?$')


@Framework.setting(section='i18n', name='localedirs')
def get_i18n_localedirs():
    """ Returns the gettext locale dir. """
    return tuple()


@Framework.setting(section='i18n', name='locales')
def get_i18n_locales():
    """ Returns the the locales actually used. """
    return None


@Framework.setting(section='i18n', name='default_locale')
def get_i18n_default_locale():
    """ Returns the fallback language to use if the user gives no indication
    towards his preference (throught the request or anything). """
    return None


@Framework.setting(section='i18n', name='locale_negotiator')
def get_i18n_locale_negotiatior():
    """ Returns the language negotiator, which is a function that takes the
    current request as well as a list of available languages and returns the
    angauge that should be used based on that information.

    """
    return default_locale_negotiator


def default_locale_negotiator(locales, request):
    """ The default locale negotiator.

    Will select one of the given locales by:

    1. Examining the 'locale' cookie which will be preferred if the
       language in the cookie actually exists

    2. Selecting the best match from the accept_language header

    If no match can be found, None is returned. It is the job of caller to
    deal with that (possibly getting :meth:`get_i18n_default_locale`).

    """
    user_locale = request.cookies.get('locale')

    if user_locale in locales:
        return user_locale

    return request.accept_language.best_match(locales)


def pofiles(localedir):
    """ Takes the given directory and yields the language and the path of
    all pofiles found under ``*/LC_MESSAGES/*.po``.

    """

    assert os.path.isdir(localedir), "Locale '{}' not found".format(localedir)

    for path in os.listdir(localedir):
        subpath = os.path.join(localedir, path, 'LC_MESSAGES')

        if not os.path.isdir(subpath):
            continue

        for pofile in glob.glob(os.path.join(subpath, '*.po')):
            yield path, pofile


@lru_cache(maxsize=32)
def compile_translation(pofile_path):
    log.info("Compiling pofile {}".format(pofile_path))

    mofile = BytesIO()
    mofile.write(polib.pofile(pofile_path).to_binary())
    mofile.seek(0)

    return gettext.GNUTranslations(mofile)


def get_translations(localedirs):
    """ Takes the given gettext locale directories and loads the po files
    found. The first found po file is assumed to be the main
    translations file (and should for performance reasons contain most of the
    translations). The other po files are added as fallbacks.

    The pofiles are compiled on-the-fly, using polib. This makes mofiles
    unnecessary.

    Returns a dictionary with the keys being languages and the values being
    GNUTranslations instances.

    """

    result = {}

    for localedir in localedirs:
        for language, pofile_path in pofiles(localedir):

            # try to automatically fix up the language
            if '_' in language:
                code, country = language.split('_')
                language = '_'.join((code.lower(), country.upper()))
            else:
                language = language.lower()

            assert VALID_LANGUAGE_EXPRESSION.match(language), """
                make sure to use languages in the following format:
                de_CH, en_GB, de, fr - note the case!
            """

            # we need to clone each translation because we might later modify
            # them when we merge translations - when multiple applications
            # share the same process this can lead to translations from one
            # application spilling over to the translations of another
            translation = clone(compile_translation(pofile_path))

            if language not in result:
                result[language] = translation
            else:
                result[language] = merge((result[language], translation))

    return result


def wrap_translations_for_chameleon(translations):
    """ Takes the given translations and wraps them for use with Chameleon. """

    return {
        lang: ChameleonTranslate(Translator(translation))
        for lang, translation in translations.items()
    }


def add_fallback_to_tail(translation, fallback):
    """ Adds the given fallback to the given translation, if it isn't already
    found somewhere in the fallback chain.

    """

    assert translation is not fallback

    while translation._fallback is not None:
        translation = translation._fallback

        if translation is fallback:
            return

    translation.add_fallback(fallback)


def get_translation_bound_meta(meta_class, translate):
    """ Takes a wtforms Meta class and combines our translate class with
    the one provided by WTForms itself.

    """

    class TranslationBoundMeta(meta_class):

        # instruct wtforms not to cache translations, there's very little
        # performance gain and it leads to some state not being cleared when
        # the request has been handled
        cache_translations = False

        def get_translations(self, form):

            try:
                default = super().get_translations(form)
                add_fallback_to_tail(translation=default, fallback=translate)
            except FileNotFoundError:
                # if there are no locales, the file not found error
                # can safely be ignored as there are no translations if there
                # are no locales (a wtforms bug)
                assert not getattr(self, 'locales', None)
                default = translate

            # Cache for reuse in render_field.
            self._translations = default

            return default

        def render_field(self, field, render_kw):
            """ Wtforms does not actually translate labels, it simply leaves
            the translations string be. If those translation strings hit our
            templates directly, they will then be picked up by our template
            renderer.

            However, that's not always the case. For example, when rendering
            a list of options, WTForms renders each option as a field with a
            label. By doing that it coerces the translation string into a
            plain string.

            By always translating the labels before they hit the template we
            can make sure that this is not a problem. We do this by
            intercepting each render call, which thankfully is something
            wtforms does support.

            """
            if hasattr(field, 'label'):
                field.label.text = self._translations.gettext(field.label.text)

            return super().render_field(
                field, render_kw)

    return TranslationBoundMeta


def get_translation_bound_form(form_class, translate):
    """ Returns a form setup with the given translate function. """

    MetaClass = get_translation_bound_meta(form_class.Meta, translate)

    class TranslationBoundForm(form_class):

        Meta = MetaClass

    return TranslationBoundForm


def merge(translations):
    """ Takes the given translations (a list) and merges them into
    each other. The translations at the end of the list are overwritten
    by the translations at the start of the list.

    This is preferrable to adding fallbacks, as they increases the average
    complexity of the lookup function each time.

    Note that the translations are *not* copied before merging, so that means
    all existing translations are changed during this processed. To avoid
    this, clone the translation first (see :func:`clone`).

    :returns:
        The last GNUTranslations object with all other translation
        objects merged into it. The first element overrides the second and so
        on.

    """
    assert len(translations) > 1

    for current, following in pairwise(translations):
        if following is not None:
            following._catalog.update(current._catalog)

    return translations[-1]


def clone(translation):
    """ Clones the given translation, creating an independent copy. """
    clone = gettext.GNUTranslations()
    clone._catalog = translation._catalog.copy()

    return clone


class SiteLocale(object):
    """ A model representing the locale of the site.

    Use this model to enable the user to change his locale through a path.

    """

    @classmethod
    def for_path(cls, app, locale, to):
        if locale in app.locales:
            return cls(locale, to)

    def __init__(self, locale, to):
        self.locale = locale
        self.to = to

    def redirect(self):
        response = morepath.redirect(self.to)
        response.set_cookie('locale', self.locale, overwrite=True)

        return response



from onegov.core.security.identity_policy import forget
from onegov.core.security.identity_policy import remembered
from onegov.core.security.permissions import Personal
from onegov.core.security.permissions import Private
from onegov.core.security.permissions import Public
from onegov.core.security.permissions import Secret


__all__ = [
    'forget',
    'Personal',
    'Private',
    'Public',
    'remembered',
    'Secret'
]



import os
import transaction

from onegov.core.datamanager import FileDataManager


def test_file_data_manager_commit(temporary_directory):

    data = 'data'.encode('utf-8')
    path = '{}/a.txt'.format(temporary_directory)
    FileDataManager.write_file(data, path)
    manager = transaction.get()._resources[0]

    assert manager.data == data
    assert manager.path == path
    assert not os.path.exists(path)

    transaction.commit()
    assert os.path.exists(path)
    with open(path) as file:
        assert file.read() == 'data'

    os.remove(path)


def test_file_data_manager_abort(temporary_directory):

    data = 'data'.encode('utf-8')
    path = '{}/b.txt'.format(temporary_directory)
    FileDataManager.write_file(data, path)
    manager = transaction.get()._resources[0]

    assert manager.data == data
    assert manager.path == path
    assert not os.path.exists(path)

    transaction.abort()
    assert not os.path.exists(path)



from datetime import date, datetime
from onegov.core.converters import extended_date_converter
from onegov.core.converters import datetime_converter
from onegov.core.converters import uuid_converter
from uuid import UUID


def test_date_converter():
    converter = extended_date_converter

    assert converter.encode(None) == ['']
    assert converter.encode('') == ['']
    assert converter.encode(date(2008, 12, 30)) == ['2008-12-30']

    assert converter.decode(['']) is None
    assert converter.decode([None]) is None
    assert converter.decode(['2008-12-30']) == date(2008, 12, 30)


def test_datetime_converter():
    converter = datetime_converter

    assert converter.encode(None) == ['']
    assert converter.encode('') == ['']
    assert converter.encode(datetime(2008, 12, 30)) == ['2008-12-30T00:00:00']

    assert converter.decode(['']) is None
    assert converter.decode([None]) is None
    assert converter.decode(['2008-12-30T12:34:56']) == datetime(
        2008, 12, 30, 12, 34, 56)


def test_uuid_converter():
    converter = uuid_converter

    assert converter.encode(None) == ['']
    assert converter.encode('') == ['']
    assert converter.encode(UUID('930a8bf4-e532-4b39-bf64-bd05e81acf01')) == \
        ['930a8bf4e5324b39bf64bd05e81acf01']

    assert converter.decode(['']) is None
    assert converter.decode([None]) is None
    assert converter.decode(['930a8bf4-e532-4b39-bf64-bd05e81acf01']) == \
        UUID('930a8bf4e5324b39bf64bd05e81acf01')



import morepath

from onegov.core.framework import Framework
from onegov.core.theme import get_filename
from webtest import TestApp as Client


class MockTheme(object):

    def __init__(self, name, version, result=''):
        self.name = name
        self.version = version
        self.result = result
        self.default_options = {}

    def compile(self, options):
        return self.result.format(**options)


def test_get_filename():
    theme = MockTheme('test', '1.0')

    assert get_filename(theme).endswith('.css')
    assert get_filename(theme) == get_filename(theme)
    assert get_filename(theme) != get_filename(MockTheme('test', '1.1'))
    assert get_filename(theme) != get_filename(MockTheme('another', '1.0'))
    assert get_filename(theme, {'x': 1}) == get_filename(theme, {'x': 1})
    assert get_filename(theme, {'x': 1}) != get_filename(theme)
    assert get_filename(theme, {'x': 1}) != get_filename(theme, {'x': 2})


def test_theme_application(temporary_directory):

    class App(Framework):
        theme_options = {
            'color': 'red'
        }

    @App.path('/')
    class Model(object):
        pass

    @App.setting(section='core', name='theme')
    def get_theme():
        return MockTheme('test', '1.0', 'body {{ background: {color} }}')

    @App.view(model=Model)
    def view_file(self, request):
        return request.theme_link

    import onegov.core
    import more.transaction
    import more.webassets
    morepath.scan(more.transaction)
    morepath.scan(more.webassets)
    morepath.scan(onegov.core)
    morepath.commit(App)

    app = App()
    app.configure_application(
        filestorage='fs.osfs.OSFS',
        filestorage_options={
            'root_path': temporary_directory
        }
    )
    app.namespace = 'tests'
    app.set_application_id('tests/foo')

    client = Client(app)
    theme_url = client.get('/').text

    assert 'background: red' in client.get(theme_url).text

    app.namespace = 'tests'
    app.set_application_id('tests/bar')

    theme_url = client.get('/').text
    assert 'background: red' in client.get(theme_url).text

    app.theme_options = {
        'color': 'blue'
    }

    new_theme_url = client.get('/').text
    assert theme_url != new_theme_url



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task
from onegov.core.orm import Base, find_models
from onegov.core.orm.abstract import Associable


@upgrade_task('Drop primary key from associated tables')
def drop_primary_key_from_associated_tables(context):
    bases = set()

    for cls in find_models(Base, lambda cls: issubclass(cls, Associable)):
        bases.add(cls.association_base())

    for base in bases:
        for link in base.registered_links.values():
            if context.has_table(link.table.name):

                command = """
                    ALTER TABLE {table} DROP CONSTRAINT IF EXISTS {table}_pkey
                """.format(table=link.table.name)

                context.operations.execute(command)



import collections
import morepath

from cached_property import cached_property
from datetime import timedelta
from onegov.core.cache import lru_cache
from itsdangerous import (
    BadSignature,
    SignatureExpired,
    TimestampSigner,
    URLSafeSerializer,
    URLSafeTimedSerializer
)
from more.webassets.core import IncludeRequest
from morepath.authentication import NO_IDENTITY
from onegov.core import utils
from onegov.core.crypto import random_token
from webob.exc import HTTPForbidden
from wtforms.csrf.session import SessionCSRF


Message = collections.namedtuple('Message', ['text', 'type'])


class ReturnToMixin(object):
    """ Provides a safe and convenient way of using return-to links.

    Return-to links are links with an added 'return-to' query parameter
    which points to the url a specific view (usually with a form) should
    return to, once all is said and done.

    There's no magic involved. If a view should honor the return-to
    paramter, it should use request.redirect instead of morepath.redirect.

    If no return-to parameter was specified, rqeuest.redirect is a
    transparent proxy to morepath.redirect.

    To create a link::

        url = request.return_to(original_url, redirect)

    To honor the paramter in a view, if present::

        return request.redirect(default_url)

    *Do not use the return-to parameter directly*. Redirect parameters
    are notorious for being used in phising attacks. By using ``return_to``
    and ``redirect`` you are kept safe from these attacks as the redirect
    url is signed and verified.

    For the same reason you should not allow the user-data for return-to
    links. Those are meant for internally generated links!

    """

    @property
    def identity_secret(self):
        raise NotImplementedError

    @property
    def redirect_signer(self):
        return URLSafeSerializer(self.identity_secret, 'return-to')

    @lru_cache(maxsize=16)
    def sign_url_for_redirect(self, url):
        return self.redirect_signer.dumps(url)

    def return_to(self, url, redirect):
        signed = self.sign_url_for_redirect(redirect)
        return utils.append_query_param(url, 'return-to', signed)

    def return_here(self, url):
        return self.return_to(url, self.url)

    def redirect(self, url):
        if 'return-to' in self.GET:
            try:
                url = self.redirect_signer.loads(self.GET['return-to'])
            except BadSignature:
                pass

        return morepath.redirect(url)


class CoreRequest(IncludeRequest, ReturnToMixin):
    """ Extends the default Morepath request with virtual host support and
    other useful methods.

    Virtual hosting might be supported by Morepath directly in the future:
    https://github.com/morepath/morepath/issues/185

    """

    @cached_property
    def identity_secret(self):
        return self.app.identity_secret

    def link_prefix(self, *args, **kwargs):
        """ Override the `link_prefix` with the application base path provided
        by onegov.server, because the default link_prefix contains the
        hostname, which is not useful in our case - we'll add the hostname
        ourselves later.

        """
        return getattr(self.app, 'application_base_path', '')

    @cached_property
    def x_vhm_host(self):
        """ Return the X_VHM_HOST variable or an empty string.

        X_VHM_HOST acts like a prefix to all links generated by Morepath.
        If this variable is not empty, it will be added in front of all
        generated urls.
        """
        return self.headers.get('X_VHM_HOST', '').rstrip('/')

    @cached_property
    def x_vhm_root(self):
        """ Return the X_VHM_ROOT variable or an empty string.

        X_VHM_ROOT is a bit more tricky than X_VHM_HOST. It tells Morepath
        where the root of the application is situated. This means that the
        value of X_VHM_ROOT must be an existing path inside of Morepath.

        We can understand this best with an example. Let's say you have a
        Morepath application that serves a blog under /blog. You now want to
        serve the blog under a separate domain, say blog.example.org.

        If we just served Morepath under blog.example.org, we'd get urls like
        this one::

            blog.example.org/blog/posts/2014-11-17-16:00

        In effect, this subdomain would be no different from example.org
        (without the blog subdomain). However, we want the root of the host to
        point to /blog.

        To do this we set X_VHM_ROOT to /blog. Morepath will then automatically
        return urls like this::

            blog.example.org/posts/2014-11-17-16:00

        """
        return self.headers.get('X_VHM_ROOT', '').rstrip('/')

    @cached_property
    def url(self):
        """ Returns the current url, taking the virtual hosting in account. """
        url = self.transform(self.path)

        if self.query_string:
            url += '?' + self.query_string

        return url

    def transform(self, url):
        """ Applies X_VHM_HOST and X_VHM_ROOT to the given url (which is
        expected to not contain a host yet!). """
        if self.x_vhm_root:
            url = '/' + utils.lchop(url, self.x_vhm_root).lstrip('/')

        if self.x_vhm_host:
            url = self.x_vhm_host + url
        else:
            url = self.host_url + url

        return url

    def link(self, *args, **kwargs):
        """ Extends the default link generating function of Morepath. """
        return self.transform(super().link(*args, **kwargs))

    def class_link(self, *args, **kwargs):
        """ Extends the default class link generating function of Morepath. """
        return self.transform(super().class_link(*args, **kwargs))

    def filestorage_link(self, path):
        """ Takes the given filestorage path and returns an url if the path
        exists. The url might point to the local server or it might point to
        somehwere else on the web.

        """

        app = self.app

        if not app.filestorage.exists(path):
            return None

        if app.filestorage.hasurl(path):
            url = app.filestorage.geturl(path)

            if not url.startswith('file://'):
                return url

        return self.link(app.modules.filestorage.FilestorageFile(path))

    @cached_property
    def theme_link(self):
        """ Returns the link to the current theme. Computed once per request.

        The theme is automatically compiled and stored if it doesn't exist yet,
        or if it is outdated.

        """
        theme = self.app.settings.core.theme
        assert theme is not None, "Do not call if no theme is used"

        force = self.app.always_compile_theme or (
            self.app.allow_shift_f5_compile and
            self.headers.get('cache-control') == 'no-cache' and
            self.headers.get('x-requested-with') != 'XMLHttpRequest')

        filename = self.app.modules.theme.compile(
            self.app.themestorage, theme, self.app.theme_options,
            force=force
        )

        return self.link(self.app.modules.theme.ThemeFile(filename))

    @cached_property
    def browser_session(self):
        """ Returns a browser_session bound to the request. Works via cookies,
        so requests without cookies won't be able to use the browser_session.

        The browser session is bound to the application (by id), so no session
        data is shared between the applications.

        If no data is written to the browser_session, no session_id cookie
        is created.

        """

        if 'session_id' in self.cookies:
            session_id = self.app.unsign(self.cookies['session_id'])
            session_id = session_id or random_token()
        else:
            session_id = random_token()

        def on_dirty(namespace, token):
            self.cookies['session_id'] = self.app.sign(token)

            @self.after
            def store_session(response):
                response.set_cookie(
                    'session_id',
                    self.cookies['session_id'],
                    secure=self.app.identity_secure,
                    httponly=True
                )

        return self.app.modules.browser_session.BrowserSession(
            namespace=self.app.application_id,
            token=session_id,
            cache=self.app.session_cache,
            on_dirty=on_dirty
        )

    def get_form(self, form_class, i18n_support=True, csrf_support=True,
                 data=None, model=None):
        """ Returns an instance of the given form class, set up with the
        correct translator and with CSRF protection enabled (the latter
        doesn't work yet).

        Form classes passed to this function (or defined through the
        ``App.form`` directive) may define a ``on_request`` method, which
        is called after the request has been bound to the form and before
        the view function is called.

        """
        meta = {}

        if i18n_support:
            translate = self.get_translate(for_chameleon=False)
            form_class = self.app.modules.i18n.get_translation_bound_form(
                form_class, translate)

            meta['locales'] = [self.locale, 'en'] if self.locale else []

        if csrf_support:
            meta['csrf'] = True
            meta['csrf_context'] = self.browser_session
            meta['csrf_class'] = SessionCSRF
            meta['csrf_secret'] = self.app.csrf_secret.encode('utf-8')
            meta['csrf_time_limit'] = timedelta(
                seconds=self.app.csrf_time_limit)

        form = form_class(self.POST, meta=meta, data=data)

        assert not hasattr(form, 'request')
        form.request = self
        form.model = model

        if hasattr(form, 'on_request'):
            form.on_request()

        return form

    def translate(self, text):
        """ Transalates the given text, if it's a translatable text. """

        if not hasattr(text, 'domain'):
            return text

        return self.translator(text)

    @cached_property
    def translator(self):
        """ Returns the translate function for basic string translations. """
        translator = self.get_translate()
        if translator:
            return lambda text: text.interpolate(translator.gettext(text))

        return lambda text: text.interpolate(text)

    @cached_property
    def default_locale(self):
        """ Returns the default locale. """
        return self.app.settings.i18n.default_locale

    @cached_property
    def locale(self):
        """ Returns the current locale of this request. """
        settings = self.app.settings

        locale = settings.i18n.locale_negotiator(self.app.locales, self)

        return locale or settings.i18n.default_locale

    @cached_property
    def html_lang(self):
        """ The language code for the html tag. """
        return self.locale and self.locale.replace('_', '-') or ''

    def get_translate(self, for_chameleon=False):
        """ Returns the translate method to the given request, or None
        if no such method is availabe.

        :for_chameleon:
            True if the translate instance is used for chameleon (which is
            special).

        """

        if not self.app.locales:
            return None

        if for_chameleon:
            return self.app.chameleon_translations.get(self.locale)
        else:
            return self.app.translations.get(self.locale)

    def message(self, text, type):
        """ Adds a message with the given type to the messages list. This
        messages list may then be displayed by an application building on
        onegov.core.

        For example:

            http://foundation.zurb.com/docs/components/alert_boxes.html

        Four default types are defined on the request for easier use:

        :meth:`success`
        :meth:`warning`
        :meth:`info`
        :meth:`alert`

        The messages are stored with the session and to display them, the
        template using the messages should call :meth:`consume_messages`.

        """
        if not self.browser_session.has('messages'):
            self.browser_session.messages = [Message(text, type)]
        else:
            # this is a bit akward, but I don't see an easy way for this atm.
            # (otoh, usually there's going to be one message only)
            self.browser_session.messages = self.browser_session.messages + [
                Message(text, type)
            ]

    def consume_messages(self):
        """ Returns the messages, removing them from the session in the
        process. Call only if you can be reasonably sure that the user
        will see the messages.

        """
        if self.browser_session.has('messages'):
            yield from self.browser_session.messages
            del self.browser_session.messages

    def success(self, text):
        """ Adds a success message. """
        self.message(text, 'success')

    def warning(self, text):
        """ Adds a warning message. """
        self.message(text, 'warning')

    def info(self, text):
        """ Adds an info message. """
        self.message(text, 'info')

    def alert(self, text):
        """ Adds an alert message. """
        self.message(text, 'alert')

    @cached_property
    def is_logged_in(self):
        """ Returns True if the current request is logged in at all. """
        return self.identity is not NO_IDENTITY

    def has_permission(self, model, permission):
        """ Returns True if the current user has the given permission on the
        given model.

        """
        if permission is None:
            return True

        return self.app._permits(self.identity, model, permission)

    def has_access_to_url(self, url):
        """ Returns true if the current user has access to the given url.

        The domain part of the url is completely ignored. This method should
        only be used if you have no other choice. Loading the object by
        url first is slower than if you can get the object otherwise.

        The initial use-case for this function is the to parameter in the
        login view. If the to-url is accessible anyway, we skip the login
        view.

        If we can't find a view for the url, a KeyError is thrown.

        """
        obj, view_name = self.app.object_by_path(url, with_view_name=True)

        if obj is None:
            raise KeyError("Could not find view for '{}'".format(url))

        permission = self.app.permission_by_view(obj, view_name)
        return self.has_permission(obj, permission)

    def exclude_invisible(self, models):
        """ Excludes models invisble to the current user from the list. """
        return [m for m in models if self.is_visible(m)]

    def is_visible(self, model):
        """ Returns True if the given model is visible to the current user.
        This is basically an alias for :meth:`CoreRequest.is_public`. It exists
        because it is easier to understand than ``is_public``.

        """
        return self.has_permission(model, self.app.modules.security.Public)

    def is_public(self, model):
        """ Returns True if the current user has the Public permission for
        the given model.

        """
        return self.has_permission(model, self.app.modules.security.Public)

    def is_personal(self, model):
        """ Returns True if the current user has the Personal permission for
        the given model.

        """
        return self.has_permission(model, self.app.modules.security.Personal)

    def is_private(self, model):
        """ Returns True if the current user has the Private permission for
        the given model.

        """
        return self.has_permission(model, self.app.modules.security.Private)

    def is_secret(self, model):
        """ Returns True if the current user has the Secret permission for
        the given model.

        """
        return self.has_permission(model, self.app.modules.security.Secret)

    @cached_property
    def current_role(self):
        """ Returns the user-role of the current request, if logged in.
        Otherwise, None is returned.

        """
        return self.is_logged_in and self.identity.role or None

    def has_role(self, *roles):
        """ Returns true if the current user has any of the given roles. """

        assert roles and all(roles)
        return self.current_role in roles

    @cached_property
    def csrf_salt(self):
        if not self.browser_session.has('csrf_salt'):
            self.browser_session['csrf_salt'] = random_token()

        return self.browser_session['csrf_salt']

    def new_csrf_token(self, salt=None):
        """ Returns a new CSRF token. A CSRF token can be verified
        using :meth:`is_valid_csrf_token`.

        Note that forms do their own CSRF protection. This is meant
        for CSRF protection outside of forms.

        onegov.core uses the Synchronizer Token Pattern for CSRF protection:
        `<https://www.owasp.org/index.php/\
        Cross-Site_Request_Forgery_%28CSRF%29_Prevention_Cheat_Sheet>`_

        New CSRF tokens are signed usign a secret attached to the session (but
        not sent out to the user). Clients have to return the CSRF token they
        are given. The token has to match the secret, which the client doesn't
        know. So an attacker would have to get access to both the cookie and
        the html source to be able to forge a request.

        Since cookies are marked as HTTP only (no javascript access), this
        even prevents CSRF attack combined with XSS.

        """
        # no csrf tokens for anonymous users (there's not really a point
        # to doing this)
        if not self.is_logged_in:
            return ''

        assert salt or self.csrf_salt
        salt = salt or self.csrf_salt

        # use app.identity_secret here, because that's being used for
        # more.itsdangerous, which uses the same algorithm
        signer = TimestampSigner(self.identity_secret, salt=salt)

        return signer.sign(random_token())

    def assert_valid_csrf_token(self, signed_value=None, salt=None):
        """ Validates the given CSRF token and returns if it was
        created by :meth:`new_csrf_token`. If there's a mismatch, a 403 is
        raised.

        If no signed_value is passed, it is taken from
        request.params.get('csrf-token').

        """
        signed_value = signed_value or self.params.get('csrf-token')
        salt = salt or self.csrf_salt

        if not signed_value:
            raise HTTPForbidden()

        if not salt:
            raise HTTPForbidden()

        signer = TimestampSigner(self.identity_secret, salt=salt)
        try:
            signer.unsign(signed_value, max_age=self.app.csrf_time_limit)
        except (SignatureExpired, BadSignature):
            raise HTTPForbidden()

    def new_url_safe_token(self, data, salt=None):
        """ Returns a new URL safe token. A token can be deserialized
        using :meth:`load_url_safe_token`.

        """
        serializer = URLSafeTimedSerializer(self.identity_secret)
        return serializer.dumps(data, salt=salt)

    def load_url_safe_token(self, data, salt=None, max_age=3600):
        """ Deserialize a token created by :meth:`new_url_safe_token`.

        If the token is invalid, None is returned.

        """
        serializer = URLSafeTimedSerializer(self.identity_secret)
        try:
            return serializer.loads(data, salt=salt, max_age=max_age)
        except (SignatureExpired, BadSignature):
            return None



""" Offers tools to deal with csv (and xls, xlsx) files. """

import codecs
import re
import sys
import tempfile
import xlrd

from collections import namedtuple, OrderedDict
from csv import DictWriter, Sniffer, QUOTE_ALL
from csv import Error as CsvError
from csv import reader as csv_reader
from csv import writer as csv_writer
from editdistance import eval as distance
from onegov.core.cache import lru_cache
from io import BytesIO, StringIO
from itertools import permutations
from onegov.core import errors
from ordered_set import OrderedSet
from unidecode import unidecode
from xlsxwriter.workbook import Workbook


VALID_CSV_DELIMITERS = {',', ';', '\t'}
WHITESPACE = re.compile(r'\s+')


class CSVFile(object):
    """ Provides access to a csv file.

    :param csvfile:
        The csv file to be accessed. Must be an open file (not a poth), opened
        in binary mode. For example::

            with open(path, 'rb') as f:
                csv = CSVFile(f)

    :param expected_headers:
        The expected headers if known. Expected headers are headers which
        *must* exist in the CSV file. There may be additional headers.

        If the headers are slightly misspelled, a matching algorithm tries to
        guess the correct header, without accidentally matching the wrong
        headers.

        See :func:`match_headers` for more information.

        If the no expected_headers are passed, no checks are done, but the
        headers are still available. Headers matching is useful if a user
        provides the CSV and it might be wrong.

        If it is impossible for misspellings to occurr, the expected headers
        don't have to be specified.

    :param dialect:
        The CSV dialect to expect. By default, the dialect will be guessed
        using Python's heuristic.

    :param encoding:
        The CSV encoding to expect. By default, the encoding will be guessed
        and will either be UTF-8 or CP1252.

    :param rename_duplicate_column_names:
        It is possible to rename duplicate column names to deal with super
        wacky files. If this option is set and a duplicate header is found,
        a suffix is appended to the column name rather than throwing a
        DuplicateColumnNamesError.

    :param rowtype:
        An alternative rowtype for the resulting rows. This should be a
        callable that receives a `rownumber` key/value and all the other
        keys/values found in the csv. The keys are normalized and are valid
        Python identifiers usable as attribute names.

        Defaults to a namedtuple created using the found headers.

    Once the csv file is open, the records can be acceessed as follows::

        with open(path, 'rb') as f:
            csv = CSVFile(f)

            for line in csv.lines:
                csv.my_field  # access the column with the 'my_field' header

    """

    def __init__(self, csvfile, expected_headers=None, dialect=None,
                 encoding=None, rename_duplicate_column_names=False,
                 rowtype=None):

        # guess the encoding if not already provided
        encoding = encoding or detect_encoding(csvfile)
        if encoding is None:
            raise errors.InvalidFormatError()

        self.csvfile = codecs.getreader(encoding)(csvfile)

        # sniff the dialect if not already provided
        try:
            self.csvfile.seek(0)
            self.dialect = dialect or sniff_dialect(self.csvfile.read(1024))
        except (CsvError, errors.InvalidFormatError):
            self.csvfile.seek(0)
            self.dialect = sniff_dialect(self.csvfile.read())

        # match the headers
        self.csvfile.seek(0)

        # if no expected headers expect, we just take what we can get
        headers = parse_header(
            self.csvfile.readline(),
            self.dialect,
            rename_duplicate_column_names
        )
        expected_headers = expected_headers or headers

        self.headers = OrderedDict(
            (h, c) for c, h in enumerate(match_headers(
                headers=headers,
                expected=expected_headers
            ))
        )

        # create an output type
        assert 'rownumber' not in expected_headers, """
            rownumber can't be used as a header
        """

        self.rowtype = rowtype or namedtuple(
            "CSVFileRow", ['rownumber'] + list(
                self.as_valid_identifier(k)
                for k in self.headers.keys()
            )
        )

    @lru_cache(maxsize=128)
    def as_valid_identifier(self, value):
        result = normalize_header(value)
        for invalid in '- .%/':
            result = result.replace(invalid, '_')
        while result and result[0] in '_0123456789':
            result = result[1:]
        return result

    @property
    def lines(self):
        self.csvfile.seek(0)

        encountered_empty_line = False

        for ix, line in enumerate(csv_reader(self.csvfile, self.dialect)):

            # raise an empty line error if we found one somewhere in the
            # middle -> at the end they don't count
            if not line:
                encountered_empty_line = True
                continue

            if line and encountered_empty_line:
                raise errors.EmptyLineInFileError()

            # the first line is the header
            if ix == 0:
                continue

            yield self.rowtype(
                rownumber=ix + 1,  # row numbers are for customers, not coders
                **{
                    self.as_valid_identifier(header): line[column].strip()
                    for header, column in self.headers.items()
                }
            )


def detect_encoding(csvfile):
    """ Since encoding detection is hard to get right (and work correctly
    every time), we limit ourselves here to UTF-8 or CP1252, whichever works
    first. CP1252 is basically the csv format you get if you use windows and
    excel and it is a superset of ISO-8859-1/LATIN1.

    """
    csvfile.seek(0)

    try:
        for line in csvfile.readlines():
            line.decode('utf-8')

        return 'utf-8'
    except UnicodeDecodeError:
        return 'cp1252'


def sniff_dialect(csv):
    """ Takes the given csv string and returns the dialect or raises an error.
    Works just like Python's built in sniffer, just that it is a bit more
    conservative and doesn't just accept any kind of character as csv
    delimiter.

    """
    if not csv:
        raise errors.EmptyFileError()

    try:
        dialect = Sniffer().sniff(csv, VALID_CSV_DELIMITERS)
    except CsvError:

        # sometimes we can get away with an extra pass just over the first line
        # (the header tends to contain fewer special cases)
        if '\n' in csv:
            return sniff_dialect(csv[:csv.find('\n')])

        raise errors.InvalidFormatError()

    return dialect


def normalize_header(header):
    """ Normalizes a header value to be as uniform as possible.

    This includes:
        * stripping the whitespace around it
        * lowercasing everything
        * transliterating unicode (e.g. 'ä' becomes 'a')
        * removing duplicate whitespace inside it
    """

    header = header.strip()
    header = header.lower()
    header = unidecode(header)
    header = WHITESPACE.sub(' ', header)

    return header


def convert_xls_to_csv(xls, sheet_name=None):
    """ Takes an XLS/XLSX file and returns a csv file using the given worksheet
    name or the first worksheet found.

    """

    xls.seek(0)

    try:
        excel = xlrd.open_workbook(file_contents=xls.read())
    except UnicodeDecodeError:
        raise IOError(
            "Could not read excel file, "
            "be sure to open the file in binary mode!"
        )

    if sheet_name:
        try:
            sheet = excel.sheet_by_name(sheet_name)
        except xlrd.XLRDError:
            raise IOError(
                "Could not find the given sheet in this excel file!"
            )
    else:
        sheet = excel.sheet_by_index(0)

    # XXX we want the output to be bytes encoded, which is not possible using
    # python's csv module. I'm sure there's a clever way of doing this by
    # using a custom StringIO class which encodes on the fly, but I haven't
    # looked into it yet. So for now a conversion is done with some memory
    # overhead.
    text_output = StringIO()
    writecsv = csv_writer(text_output, quoting=QUOTE_ALL)

    for rownum in range(0, sheet.nrows):
        values = []

        for cell in sheet.row(rownum):

            if cell.ctype == xlrd.XL_CELL_TEXT:
                value = cell.value
            elif cell.ctype in (xlrd.XL_CELL_EMPTY, xlrd.XL_CELL_BLANK):
                value = ''
            elif cell.ctype == xlrd.XL_CELL_NUMBER:
                if cell.value.is_integer():
                    value = str(int(cell.value))
                else:
                    value = str(cell.value)
            elif cell.ctype == xlrd.XL_CELL_DATE:
                value = xlrd.xldate_as_tuple(cell.value, excel.datemode)
                value = value.isoformat()
            else:
                raise NotImplementedError

            values.append(value)

        writecsv.writerow(values)

    text_output.seek(0)
    output = BytesIO()

    for line in text_output.readlines():
        output.write(line.encode('utf-8'))

    return output


def get_keys_from_list_of_dicts(rows, key=None, reverse=False):
    """ Returns all keys of a list of dicts in an ordered tuple.

    If the list of dicts is irregular, the keys found in later rows are
    added at the end of the list.

    Note that the order of keys is otherwise defined by the order of the keys
    of the dictionaries. So if ordered dictionaries are used, the order is
    defined. If regular dictionaries are used, the order is undefined.

    Alternatively, a key and a reverse flag may be provided which will be
    used to order the fields. If the list of fields is specified, the key and
    the reverse flag is ignored.

    """
    fields = OrderedSet()

    for dictionary in rows:
        fields.update(dictionary.keys())

    if key:
        fields = tuple(sorted(fields, key=key, reverse=reverse))
    else:
        fields = tuple(fields)

    return fields


def convert_list_of_dicts_to_csv(rows, fields=None, key=None, reverse=False):
    """ Takes a list of dictionaries and returns a csv.

    If no fields are provided, all fields are included in the order of the keys
    of the first dict. With regular dictionaries this is random. Use an ordered
    dict or provide a list of fields to have a fixed order.

    Alternatively, a key and a reverse flag may be provided which will be
    used to order the fields. If the list of fields is specified, the key and
    the reverse flag is ignored.

    The function returns a string created in memory. Therefore this function
    is limited to small-ish datasets.

    """

    if not rows:
        return ''

    fields = fields or get_keys_from_list_of_dicts(rows, key, reverse)

    output = StringIO()
    writer = DictWriter(output, fieldnames=fields)

    writer.writeheader()

    for row in rows:
        writer.writerow({field: row.get(field, '') for field in fields})

    output.seek(0)
    return output.read()


def convert_list_of_dicts_to_xlsx(rows, fields=None, key=None, reverse=False):
    """ Takes a list of dictionaries and returns a xlsx.

    This behaves the same way as :func:`convert_list_of_dicts_to_csv`.

    """

    small_chars = set('fijlrt:,;.+i ')
    large_chars = set('GHMWQ_')

    max_width = 75

    def character_width(char):
        # those numbers have been acquired by chasing unicorns
        # and fairies in the magic forest of Excel
        #
        # tweak them as needed, but know that there's no correct answer,
        # as each excel version on each platform or os-version will render
        # the fonts used at different widths
        if char in small_chars:
            return 0.75
        elif char in large_chars:
            return 1.2
        else:
            return 1

    def estimate_width(text):
        if not text:
            return 0

        width = max(
            sum(character_width(c) for c in line)
            for line in text.splitlines()
        )

        return min(width, max_width)

    with tempfile.NamedTemporaryFile() as file:
        workbook = Workbook(file.name, options={'constant_memory': True})
        cellformat = workbook.add_format({'text_wrap': True})

        worksheet = workbook.add_worksheet()

        fields = fields or get_keys_from_list_of_dicts(rows, key, reverse)

        # write the header
        worksheet.write_row(0, 0, fields, cellformat)

        # keep track of the maximum character width
        column_widths = [estimate_width(field) for field in fields]

        def values(row):
            for ix, field in enumerate(fields):
                value = row.get(field, '')
                column_widths[ix] = max(
                    column_widths[ix],
                    estimate_width(str(value))
                )

                if isinstance(value, str):
                    value = value.replace('\r', '')

                yield value

        # write the rows
        for r, row in enumerate(rows, start=1):
            worksheet.write_row(r, 0, values(row), cellformat)

        # set the column widths
        for col, width in enumerate(column_widths):
            worksheet.set_column(col, col, width)

        workbook.close()

        file.seek(0)
        return file.read()


def parse_header(csv, dialect=None, rename_duplicate_column_names=False):
    """ Takes the first line of the given csv string and returns the headers.

    Headers are normalized (stripped and normalized) and expected to be
    unique. The dialect is sniffed, if not provided.

    :return: A list of headers in the order of appearance.

    """
    try:
        dialect = dialect or sniff_dialect(csv)
    except errors.InvalidFormatError:
        dialect = None  # let the csv reader try, it might still work
    except errors.EmptyFileError:
        return []

    headers = next(csv_reader(csv.splitlines(), dialect=dialect))
    headers = [normalize_header(h) for h in headers if h]

    if rename_duplicate_column_names:
        indexes = {}
        for i, item in enumerate(headers):
            indexes.setdefault(item, []).append(i)
        for key, value in indexes.items():
            for suffix, index in enumerate(value[1:]):
                headers[index] += '_{}'.format(suffix + 1)

    return headers


def match_headers(headers, expected):
    """ Takes a list of normalized headers and matches them up against a
    list of expected headers.

    The headers may differ from the expected headers. This function tries to
    match them up using the Levenshtein distance. It does so somewhat carefully
    by calculating a sane distance using the input.

    The passed headers are expected to be have been normalized by
    :func:`normalize_header`, since we usually will pass the result of
    :func:`parse_header` to this function.

    For example::

        match_headers(
            headers=['firstname', 'lastname'],
            expected=['last_name', 'first_name']
        )

    Results in::

        ['first_name', 'last_name']

    If no match is possible, an
    :class:`~onegov.core.errors.MissingColumnsError`, or
    :class:`~onegov.core.errors.AmbiguousColumnsError` or
    :class:`~onegov.core.errors.DuplicateColumnNamesError` error is raised.

    :return: The matched headers in the order of appearance.

    """

    # for this to work we need unique header names
    if len(headers) != len(set(headers)):
        raise errors.DuplicateColumnNamesError()

    # we calculate a 'sane' levenshtein distance by comparing the
    # the distances between all headers, permutations, as well as the lengths
    # of all expected headers. This makes sure we don't end up with matches
    # that make no sense (like ['first', 'second'] matching ['first', 'third'])
    sane_distance = getattr(sys, 'maxsize', 0) or sys.maxint

    if len(headers) > 1:
        sane_distance = min((
            sane_distance,
            min(distance(a, b) for a, b in permutations(headers, 2))
        ))

    if len(expected) > 1:
        sane_distance = min((
            sane_distance,
            min(distance(a, b) for a, b in permutations(expected, 2))
        ))

    sane_distance = min((
        sane_distance,
        min(len(c) for c in expected)
    ))

    mapping = {}
    missing = []
    ambiguous = {}

    for column in expected:
        normalized = normalize_header(column)
        distances = dict((h, distance(normalized, h)) for h in headers)

        closest = min(distances.values())

        if closest >= sane_distance:
            missing.append(column)
            continue

        matches = tuple(c for c in distances if distances[c] == closest)

        if len(matches) > 1:
            ambiguous[column] = matches
            continue

        mapping[matches[0]] = column

    if missing:
        raise errors.MissingColumnsError(missing)

    if ambiguous:
        raise errors.AmbiguousColumnsError(ambiguous)

    # return the mapping or the original header (we allow for extra headers,
    # we just don't allow for missing headers)
    return [mapping.get(h, h) for h in headers]



from sqlalchemy.types import TypeDecorator, TEXT
from sqlalchemy_utils.operators import CaseInsensitiveComparator


class LowercaseText(TypeDecorator):
    """ Text column that forces all text to be lowercase. """

    impl = TEXT
    omparator_factory = CaseInsensitiveComparator

    def process_bind_param(self, value, dialect):
        if value is not None:
            return value.lower()

        return value



from onegov.core.cli.commands import cli
from onegov.core.cli.core import (
    abort,
    command_group,
    GroupContext,
    pass_group_context
)

__all__ = [
    'abort',
    'cli',
    'command_group',
    'GroupContext',
    'pass_group_context'
]



import threading
import re
import weakref
import zope.sqlalchemy

from blinker import Signal
from onegov.core.cache import lru_cache
from sqlalchemy import create_engine, event, text
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.pool import QueuePool
from sqlalchemy.orm.query import Query
from sqlalchemy_utils.aggregates import manager as aggregates_manager


class ForceFetchQueryClass(Query):
    """ Alters the buitlin query class, ensuring that the delete query always
    fetches the data first, which is important for the bulk delete events
    to get the actual objects affected by the change.

    """

    def delete(self, synchronize_session=None):
        return super().delete('fetch')


class SessionManager(object):
    """ Holds sessions and creates schemas before binding sessions to schemas.

    Threadsafe in theory, but not tested or well thought out. No global state
    though, so two different session managers will manage different
    sessions/schemas.

    """

    # describes the accepted characters in a schema name
    _valid_schema_name = re.compile(r'^[a-z0-9_-]+$')

    # describes the prefixes with which a schema may not begin
    _invalid_prefixes = re.compile(r'^([0-9]|pg_)+')

    # holds schemas that may never be used:
    # - information_schema is a sql standard schema that's for internal use
    # - public is the default schema name which should not be used for security
    #   reasons. Someone could figure out a way to set the search_path to
    #   default and to add an application that has the right to change the
    #   public schema.
    _reserved_schemas = {'information_schema', 'public', 'extensions'}

    def __init__(self, dsn, base,
                 engine_config=None, session_config=None, pool_config=None):
        """ Configures the data source name/dsn/database url and sets up the
        connection to the database.

        :dsn:
            Database connection string including user, password, host, port
            and database name.

            See: `<http://docs.sqlalchemy.org/en/latest/core/engines.html\
            #database-urls>`_

        :base:
            Declarative base used to define the SQLAlchemy database models.

            Extra bases may be added to the session manager after __init__::

                mgr.bases.append(MyBase)

            The tables in these additional schemas are created on the schema
            alongside the primary base.

        :engine_config:
            Additional configuration passed to SQLAlchemy's `create_engine`.

            See: `<http://docs.sqlalchemy.org/en/latest/core/engines.html\
            #sqlalchemy.create_engine>`

        :session_config:
            Additional configuration passed to SQLAlchemy's sessionmaker.

            See: `<http://docs.sqlalchemy.org/en/latest/orm/session_api.html\
            #sqlalchemy.orm.session.sessionmaker>`

        :pool_config:
            Custom pool configuration to be used in lieue of the default pool.
            Only use under special circumstances.

        Note, to connect to another database you need to create a new
        SessionManager instance.

        **Postgres Extensions**

        The session manager supports the setup of postgres extensions.
        Currently those extensions are hard-coded and they are all added to the
        extensions schema. The extensions schema is then added to the search
        path of each query.

        Since extensions can only be created by superusers, this is something
        you might want to do in a separate step in deployment. We don't advise
        you to use a superuser connection for your onegov cloud deployment.

        You may therefore use the list of the required extensions below and
        create a schema 'extensions' with those extensions inside.

        **Signals**

        The session manager supports the following signals:

        ``self.on_insert``

        Called with the schema and each object when an insert happens.

        ``self.on_update``

        Called with the schema and each object when a change happens.

        ``self.on_delete``

        Called with the scheam and each object when a change happens. Note that
        because those objects basically do not exist anymore, only the
        primary keys of those objects contain actual values!

        Signals abstracts SQLAlchemy ORM events. Since those handle events
        of the Session, only things happening on sessions provided by this
        sessionmanager are caught. Raw SQL queries, other sessions and other
        outside changes to the database are not caught!

        The signals are implemented using blinker:
        `<https://pythonhosted.org/blinker/>`_

        Note that signal handler functions are by default attached as weakrefs,
        that means if you create your signal handling function inside another
        function, the signal will stop working once the outer function
        completes.

        To avoid this behavior use ``connect_via``. See examples below:

        Examples::

            mgr = SessionManager(dsn)

            # connect to the insert signal, disconnecting automatically once
            # the handler function is garbage collected
            @mgr.on_insert.connect
            def on_insert(schema, obj):
                print("Inserted {} @ {}".format(obj, schema))

            # connect to the update signal, limited to a specific schema, also
            # disconnecting automatically once the handler function falls out
            # of scope
            @mgr.on_update.connect_via('my-schema')
            def on_update(schema, obj):
                print("Updated {} @ {}".format(obj, schema))

            # connect to the delete signal for all schemas -> here, the
            # handler is strongly referenced and may be used in a closure
            from blinker import ANY
            @mgr.on_delete.connect_via(ANY, weak=False)
            def on_delete(schema, obj):
                print("Deleted {} @ {}".format(obj, schema))

        """
        assert 'postgres' in dsn, "Onegov only supports Postgres!"

        engine_config = engine_config or {}
        session_config = session_config or {}
        pool_config = pool_config or {}

        self.dsn = dsn
        self.bases = [base]
        self.created_schemas = set()
        self.current_schema = None

        self.on_insert = Signal()
        self.on_update = Signal()
        self.on_delete = Signal()

        # onegov.core creates extensions that it requires in a separate schema
        # in the future, this might become something we can configure through
        # the setuptools entry_points -> modules could advertise what they need
        # and the core would install the extensions the modules require
        self.required_extensions = {'hstore'}
        self.created_extensions = set()

        # override the isolation level in any case, we cannot allow another
        engine_config['isolation_level'] = 'SERIALIZABLE'

        if pool_config:
            engine_config.update(pool_config)
        else:
            # the default pool config should work for all onegov applications
            #
            # currently most applications use one single connection that is
            # reused between requests as well as one extra connection for each
            # instance (cronjob thread with session-based lock)
            engine_config['poolclass'] = QueuePool

            # connections which are kept open when returned to the pool
            engine_config['pool_size'] = 5

            # connections which are closed when returned (unlimited)
            engine_config['max_overflow'] = -1

        self.engine = create_engine(self.dsn, **engine_config)
        self.register_engine(self.engine)

        self.session_factory = scoped_session(
            sessionmaker(
                bind=self.engine,
                query_cls=ForceFetchQueryClass,
                **session_config
            ),
            scopefunc=self._scopefunc,
        )
        self.register_session(self.session_factory)
        self.setup_manager_instance_sharing(self.bases)

    def register_engine(self, engine):
        """ Takes the given engine and registers it with the schema
        switching mechanism. Maybe used to register external engines with
        the session manager.

        If used like this, make sure to call :meth:`bind_session` before using
        the session provided by the external engine.

        """

        @event.listens_for(engine, "before_cursor_execute")
        def _activate_schema(
            conn, cursor, statement, parameters, context, executemany
        ):
            """ Share the 'info' dictionary of Session with Connection
            objects.

            """

            # execution options have priority!
            if 'schema' in conn._execution_options:
                schema = conn._execution_options['schema']
            else:
                if 'session' in conn.info:
                    schema = conn.info['session'].info['schema']
                else:
                    schema = None

            if schema is not None:
                cursor.execute("SET search_path TO %s, extensions", (schema, ))

    def register_session(self, session):
        """ Takes the given session and registers it with zope.sqlalchemy and
        various orm events.

        """

        signals = (
            (self.on_insert, lambda session: session.new),
            (self.on_update, lambda session: session.dirty),
            (self.on_delete, lambda session: session.deleted)
        )

        # SQLAlchemy-Utils' aggregates decorator doesn't work correctly
        # with our sessions - we need to setup the appropriate event handler
        # manually. There's a test that makes sure that this is not done
        # twice. The cause of this has to be investigated still.
        event.listen(
            target=session,
            identifier='after_flush',
            fn=aggregates_manager.construct_aggregate_queries
        )

        # This check we need in any case, since aggregates don't work with
        # bulk udpdates/deletes, which is something we make sure can't
        # happen by accident (it'll lead to hard to debug errors)
        cache_size = min(len(aggregates_manager.generator_registry), 32)

        @lru_cache(cache_size)
        def prevent_bulk_changes_on_aggregate_modules(module_class):
            for registered in aggregates_manager.generator_registry:
                assert not issubclass(module_class, registered), """
                    Bulk queries run on models that use sqlalchemy-utils
                    aggregates won't lead to a proper update. It's impossible
                    to have both aggregates and bulk updates/deletes.
                """

        @event.listens_for(session, 'after_flush')
        def on_after_flush(session, flush_context):
            for signal, get_objects in signals:
                if signal.receivers:
                    for obj in get_objects(session):
                        signal.send(self.current_schema, obj=obj)

        @event.listens_for(session, 'after_bulk_update')
        def on_after_bulk_update(update_context):
            prevent_bulk_changes_on_aggregate_modules(
                update_context.mapper.class_)

            if self.on_update.receivers:
                assert hasattr(update_context, 'matched_objects'), """
                    Bulk queries which use synchronize_session=False or
                    synchronize_session='fetch' cannot be supported because
                    it is unclear what rows were affected. Manually update
                    values instead (even though it's way slower). There's no
                    better solution at the moment.
                """

                for obj in update_context.matched_objects:
                    self.on_update.send(self.current_schema, obj=obj)

        @event.listens_for(session, 'after_bulk_delete')
        def on_after_bulk_delete(delete_context):
            prevent_bulk_changes_on_aggregate_modules(
                delete_context.mapper.class_)

            if self.on_delete.receivers:
                for row in delete_context.matched_rows:
                    obj = delete_context.mapper.class_(**{
                        c.name: v for c, v
                        in zip(delete_context.mapper.primary_key, row)
                    })
                    self.on_delete.send(self.current_schema, obj=obj)

        zope.sqlalchemy.register(session)

    def setup_manager_instance_sharing(self, bases):
        """ Shares this instance with all mapped classes of all bases.

        This enables each model to look up the current schema, the current
        locale and so on.

        This is a rather complicated way to make sure we don't use a global
        variable. One might argue that this would warrant an exception
        to the rule...

        """

        # the instance is shared with all classes, loaded or created
        def share_with_class(target, *args, **kwargs):
            target.session_manager = weakref.proxy(self)

        # and it is shared with all query types, at which point there is
        # no type available yet.
        def share_with_query(query):
            session_manager = weakref.proxy(self)

            for desc in query.column_descriptions:
                desc['type'].session_manager = session_manager

        event.listen(Query, 'before_compile', share_with_query, retval=False)

        for base in bases:
            if base:
                for cls in base._decl_class_registry.values():
                    if isinstance(cls, base.__class__):
                        event.listen(cls, 'init', share_with_class)
                        event.listen(cls, 'load', share_with_class)

    def set_locale(self, default_locale, current_locale):
        """ Sets the default locale and the current locale so it may be
        shared with the translated ORM columns.

        Note that the locales may be NONE. It is up to the consumer of these
        settings to assert their existence.

        """

        self.default_locale = default_locale
        self.current_locale = current_locale

    def _scopefunc(self):
        """ Returns the scope of the scoped_session used to create new
        sessions. Relies on self.current_schema being set before the
        session is created.

        This function is as internal as they come and it exists only because
        we otherwise would have to create different session factories for each
        schema.

        """
        return (threading.current_thread(), self.current_schema)

    def dispose(self):
        """ Closes the connection to the server and cleans up. This only needed
        for testing.

        """
        self.engine.raw_connection().invalidate()
        self.engine.dispose()

    def is_valid_schema(self, schema):
        """ Returns True if the given schema looks valid enough to be created
        on the database with no danger of SQL injection or other unwanted
        sideeffects.

        """
        if not schema:
            return False

        if schema in self._reserved_schemas:
            return False

        if self._invalid_prefixes.match(schema):
            return False

        # only one consecutive '-' is allowed (-- constitues a comment)
        if '--' in schema:
            return False

        return self._valid_schema_name.match(schema) and True or False

    def set_current_schema(self, schema):
        """ Sets the current schema in use. The current schema is then used
        to bind the session to a schema. Note that this can't be done
        in a functional way. We need the current schema to generate a new
        scope.

        I would very much prefer to bind this to the session but this is not
        at all straight-forward with SQLAlchemy.

        I tried a solution like `this one <https://bitbucket.org/zzzeek/\
        sqlalchemy/wiki/UsageRecipes/SessionModifiedSQL>`_, but it's not good
        enough, because it still relies on some kind of global stage, even if
        it's set locally.

        Ideally a value could be bound to the session and an event would
        trigger every time the connection is used with that session. We
        could then set the schema on the connection every time that happens.

        For now, the global option is okay though, because in practice we
        only set the schema once per request and we don't do threading
        anyway.

        """
        assert self.is_valid_schema(schema)

        self.current_schema = schema
        self.ensure_schema_exists(schema)

    def create_schema(self, schema, validate=True):
        """ Creates the given schema. If said schema exists, expect this
        method to throw an error. If you use :meth:`set_current_schema`,
        this is invoked automatically if needed. So you usually shouldn't
        have to care about this function.

        """

        if validate:
            assert self.is_valid_schema(schema)

        # psycopg2 doesn't know how to correctly build a CREATE
        # SCHEMA statement, so we need to do it manually.
        # self.is_valid_schema should've checked that no sql
        # injections are possible.
        #
        # this is the *only* place where this happens - if anyone
        # knows how to do this using sqlalchemy/psycopg2, come forward!
        conn = self.engine.execution_options(schema=None)
        conn.execute('CREATE SCHEMA "{}"'.format(schema))
        conn.execute('COMMIT')

    def create_schema_if_not_exists(self, schema, validate=True):
        """ Creates the given schema if it doesn't exist yet. """
        if not self.is_schema_found_on_database(schema):
            self.create_schema(schema, validate)

    def bind_session(self, session):
        """ Bind the session to the current schema. """
        session.info['schema'] = self.current_schema
        session.connection().info['session'] = weakref.proxy(session)

        return session

    def session(self):
        """ Returns a new session or an existing session. Sessions with
        different schemas are kept independent, though they might reuse
        each others connections.

        This means that a session retrieved thusly::

            mgr = SessionManager('postgres://...')
            mgr.set_current_schema('foo')
            session = mgr.session()

        Will not see objects attached to this session::

            mgr.set_current_schema('bar')
            session = mgr.session()

        """

        return self.bind_session(self.session_factory())

    def list_schemas(self, limit_to_namespace=None):
        """ Returns a list containing *all* schemas defined in the current
        database.

        """
        conn = self.engine.execution_options(schema=None)
        query = text("""
            SELECT schema_name
            FROM information_schema.schemata
            ORDER BY schema_name
        """)

        if limit_to_namespace is not None:
            return [
                r[0] for r in conn.execute(query).fetchall()
                if r[0].startswith(limit_to_namespace + '-')
            ]
        else:
            return [r[0] for r in conn.execute(query).fetchall()]

    def is_schema_found_on_database(self, schema):
        """ Returns True if the given schema exists on the database. """

        conn = self.engine.execution_options(schema=None)
        result = conn.execute(text(
            "SELECT EXISTS(SELECT 1 FROM information_schema.schemata "
            "WHERE schema_name = :schema)"
        ), schema=schema)

        return result.first()[0]

    def create_required_extensions(self):
        """ Creates the required extensions once per lifetime of the manager.

        """
        if self.required_extensions != self.created_extensions:

            # extensions are a all added to a shared schema (a reserved schema)
            self.create_schema_if_not_exists('extensions', validate=False)

            conn = self.engine.execution_options(schema='extensions')
            for ext in self.required_extensions - self.created_extensions:
                conn.execute(
                    'CREATE EXTENSION IF NOT EXISTS "{}" '
                    'SCHEMA "extensions"'.format(ext)
                )
                conn.execute('COMMIT')
                self.created_extensions.add(ext)

    def ensure_schema_exists(self, schema):
        """ Makes sure the schema exists on the database. If it doesn't, it
        is created.

        """
        assert self.engine is not None

        if schema not in self.created_schemas:

            # XXX circular dependencies
            from onegov.core import upgrade

            # this is important because CREATE SCHEMA is done in a possibly
            # dangerous way!
            assert self.is_valid_schema(schema)

            # setup the extensions right before we activate our first schema
            self.create_required_extensions()
            self.create_schema_if_not_exists(schema)

            conn = self.engine.execution_options(schema=schema)
            declared_classes = set()

            try:
                for base in self.bases:
                    base.metadata.schema = schema
                    base.metadata.create_all(conn)

                    declared_classes.update(base._decl_class_registry.values())

                conn.execute('COMMIT')
            finally:
                # reset the schema on the global base variable - this state
                # sticks around otherwise and haunts us in the tests
                for base in self.bases:
                    base.metadata.schema = None

            # if we have an upgrade state table, we want to prefill it with
            # all the current modules/tasks, to get the correct initial update
            # state (see https://github.com/OneGov/onegov.core/issues/8)
            #
            # we usually want to do that, but during testing, those upgrade
            # state classes may not exist, that's why we check
            if upgrade.UpgradeState in declared_classes:

                # we use a special session that is not registered with the
                # zope transaction extension -> we do this during a request
                # and we don't want the whole request to be commited, just
                # this small change.
                session = sessionmaker(bind=self.engine)()
                upgrade.register_all_modules_and_tasks(session)
                session.commit()

            self.created_schemas.add(schema)



""" Provides caching methods for onegov.core.

Onegov.core uses dogpile for caching:
`<https://dogpilecache.readthedocs.org/>`_

Unlike dogpile onegov.core does not provide a global region however.
The cache is available through the app::

    request.app.cache.set('key', 'value')

Global caches in a multi-tennant application are a security vulnerability
waiting to be discovered. Therefore we do not do that!

This means that this won't be available::

    from x import cache
    @cache.cache_on_arguments()
    def my_function():
        return 'foobar'

Syntactic sugar like this will be provided through decorators inside this
module in the future. For example, we could write one that is usable
on all morepath views::

    @App.view(...)
    @cache.view()
    def my_view():
        return '<html>...'

But no such method exists yet.

Currently there is one cache per app that never expires (though values will
eventually be discarded by memcache if the cache is full).

"""

import dill
import memcache

from fastcache import clru_cache as lru_cache  # noqa
from dogpile.cache import make_region, register_backend
from dogpile.cache.backends.memcached import MemcachedBackend as BaseBackend
from dogpile.cache.backends.memory import MemoryBackend as BaseMemoryBackend
from dogpile.cache.api import NO_VALUE
from dogpile.cache.proxy import ProxyBackend
from hashlib import sha1
from onegov.core import log


class MemcachedBackend(BaseBackend):
    """ A custom memcached backend with the following improvements:

    * Uses dill instead of pickle, allowing for a wider range of cached objs.

    """

    def _create_client(self):
        return memcache.Client(
            self.url,
            pickler=DillPickler,
            unpickler=DillPickler
        )


register_backend(
    'onegov.core.memcached',
    'onegov.core.cache',
    'MemcachedBackend'
)

dill.settings['recurse'] = True


class DillPickler(object):
    """ A python-memcached pickler that uses dill instead of the builtin
    pickle module.

    Dill is an alternative implementation that supports a wider variety
    of objects which it can pickle.

    """

    def __init__(self, file, protocol=None):
        self.file = file

    def dump(self, value):
        return dill.dump(value, self.file)

    def load(self):
        return dill.load(self.file)


class MemoryBackend(BaseMemoryBackend):
    """ A custom memory backend that uses dill to pickle values into a
    dictionary. This has the added benefit of using the same codepath as the
    memcached based backend, without actually requiring a memcached server.

    """

    def get(self, key):
        value = self._cache.get(key, NO_VALUE)
        return value if value is NO_VALUE else dill.loads(value)

    def get_multi(self, keys):
        ret = (self._cache.get(key, NO_VALUE) for key in keys)
        return [
            value if value is NO_VALUE else dill.loads(value)
            for value in ret
        ]

    def set(self, key, value):
        self._cache[key] = dill.dumps(value)

    def set_multi(self, mapping):
        for key, value in mapping.items():
            value = dill.dumps(value)
            self._cache[key] = value


register_backend(
    'onegov.core.memory',
    'onegov.core.cache',
    'MemoryBackend'
)


def prefix_key_mangler(prefix):

    prefix = prefix.encode('utf-8')

    def key_mangler(key):
        return sha1(prefix + key.encode('utf-8')).hexdigest()

    return key_mangler


def create_backend(namespace, backend, arguments={}, expiration_time=None):

    prefix = '{}:'.format(namespace)

    return make_region(key_mangler=prefix_key_mangler(prefix)).configure(
        backend,
        expiration_time=expiration_time,
        arguments=arguments,
        wrap=[IgnoreUnreachableBackend]
    )


class IgnoreUnreachableBackend(ProxyBackend):
    """ A proxy backend that logs an error when memcached is down, but keeps
    running, albeit without using any caching.

    The idea is that a memcached outage will result in a slower/degraded
    user experience, not in complete breakage.

    """

    def get(self, key):
        try:
            return self.proxied.get(key)
        except TypeError:
            log.exception("Error reading from memcached")
            return NO_VALUE

    def set(self, key, value):
        try:
            self.proxied.set(key, value)
        except TypeError:
            log.exception("Error writing to memcached")

    def delete(self, key):
        try:
            self.proxied.delete(key)
        except TypeError:
            log.exception("Error deleting from memcached")

    def get_multi(self, keys):
        try:
            return self.proxied.get_multi(keys)
        except TypeError:
            log.exception("Error reading from memcached")
            return [NO_VALUE] * len(keys)

    def set_multi(self, mapping):
        try:
            self.proxied.set_multi(mapping)
        except TypeError:
            log.exception("Error writing to memcached")

    def delete_multi(self, keys):
        try:
            self.proxied.delete_multi(keys)
        except TypeError:
            log.exception("Error deleting from memcached")



import logging
log = logging.getLogger('onegov.core')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from onegov.core.framework import Framework
from onegov.core.filestorage import get_filestorage_file # noqa

# include the filters module so they get picked up by webassets
from onegov.core import filters  # noqa

__all__ = ['Framework', 'log']



""" Contains custom converters used by onegov.town. """

import isodate
import morepath

from datetime import date, datetime
from onegov.core.framework import Framework
from onegov.core.utils import is_uuid
from time import mktime, strptime
from uuid import UUID


def extended_date_decode(s):
    """ Decodes a date string HTML5 (RFC3339) compliant."""
    if not s:
        return None

    return date.fromtimestamp(mktime(strptime(s, '%Y-%m-%d')))


def extended_date_encode(d):
    """ Encodes a date HTML5 (RFC3339) compliant. """
    if not d:
        return ''

    return d.strftime('%Y-%m-%d')


extended_date_converter = morepath.Converter(
    decode=extended_date_decode, encode=extended_date_encode
)


def uuid_decode(s):
    """ Turns a uuid string into a UUID instance. """

    return is_uuid(s) and UUID(s) or None


def uuid_encode(uuid):
    """ Turns a UUID instance into a uuid string. """
    if not uuid:
        return ''

    if isinstance(uuid, str):
        return uuid

    return uuid.hex


uuid_converter = morepath.Converter(
    decode=uuid_decode, encode=uuid_encode
)


@Framework.converter(type=UUID)
def get_default_uuid_converter():
    return uuid_converter


def bool_decode(s):
    """ Decodes a boolean. """
    return False if s == '0' or s == '' else True


def bool_encode(d):
    """ Encodes a boolean. """
    return d and '1' or '0'


bool_converter = morepath.Converter(
    decode=bool_decode, encode=bool_encode
)


@Framework.converter(type=bool)
def get_default_bool_converter():
    return bool_converter


def datetime_decode(s):
    """ Decodes a datetime. """
    return None if not s else isodate.parse_datetime(s)


def datetime_encode(d):
    """ Encodes a datetime. """
    return isodate.datetime_isoformat(d) if d else ''


datetime_converter = morepath.Converter(
    decode=datetime_decode, encode=datetime_encode
)


@Framework.converter(type=datetime)
def get_default_datetime_converter():
    return datetime_converter



"""
.. _core-commands:

Core Commands
-------------

Provides a framework for cli commands run against one ore more onegov
cloud applications.

OneGov cli commands are usually ran against a onegov.yml config file, which
may contain definitions for multiple applications. It may define multiple
application with different applicaiton classes and it may contain wildcard
applications which run the same application class, but contain multiple
tennants for each application.

To have a command run against one or many applications we use a selector to
help select the applications we want to target in a command.

In addition to selectors, onegov core cli commands provide a simple way to
write a function that takes a request and an application. This function is
then called for each application matching the selector, with a proper
request and application context already setup (with the same characteristics
as if called through an url in the browser).

Selector
--------

A selector has the form <namespace>/<id>.

That is, it consists of the namespace of the application and it's id.

For example:

    * ``/foo/bar``
    * ``/onegov_election_day/gr``
    * ``/onegov_town/govikon``

To select non-wildcard applications we can just omit the id:

    * ``/foo``
    * ``/onegov_onboarding``

Finally, to select multiple applications we can use wildcards:

    * ``/foo/*``
    * ``/onegov_election_day/*``
    * ``/*/g??``

Execution
---------

To run a supported command we provide a selector as an option::

    bin/onegov-core --select '/foo/*' subcommand

To find out what kind of selectors are available, we can simply run::

    bin/onegov-core

Which will print out a list of selector suggestions.

Registering a Selector Based Command
------------------------------------

To write a selector based command we first create a command group::

    from onegov.core.cli import command_group
    cli = command_group()

Using that command group, we can register our own commands::

    @cli.command()
    def my_click_command():
        pass

This command works like any other click command::

    import click

    @cli.command()
    @click.option('--option')
    def my_click_command(option):
        pass

Each command has the ability to influence the way selectors work. For example,
a command which creates the path that matches the selector we can use::

    @cli.command(context_settings={'creates_path': True})

By default we expect that a selector is passed. For commands which usually run
against all applications we can provide a default selector::

    @cli.command(context_settings={'default_selector': '*'})

Using the app/request context
-----------------------------

For a lot of commands the easiest approach is to have a function which is
called for each application with a request. This allows us to write commands
which behave like they were written in a view.

To do that we register a command which returns a function with the following
signature::

    def handle_command(request, app):
        pass

For example::

    @cli.command()
    def my_click_command():

        def handle_command(request, app):
            pass

        return handle_command

Setup like this, ``handle_command`` will be called with a request for each
application (and tennant). This function acts exactly like a view. Most
importantly, it does *not* require transaction commits, because like with
ordinary requests, the transaction is automatically committed if no error
occurs.

Using the app configurations directly
-------------------------------------

Sometimes we don't want to use the request/app context, or maybe we want to
setup something before receiving a request.

To do this, we use the ``pass_group_context`` decorator.

For example::

    from onegov.core.cli import pass_group_context

    @cli.command()
    @pass_group_context
    def my_click_command(group_context):

        for appcfg in group_context.appcfgs:
            # do something

This is independent of the app/request context. If we return a function, the
function is going to be called with the request and the app. If we do not, the
command ends as expected.

Limiting Selectors to a Single Instance
---------------------------------------

Sometimes we want to write commands which only run against a single
application. A good example is a command which returns 1/0 depending on the
existence of something *in* an application.

To do that, we use::

    @cli.command(context_settings={'singular': True})
    def my_click_command():
        pass

If a selector is passed which matches more than one application, the command
is not executed.

"""

import click
import inspect
import sqlalchemy
import sys

from fnmatch import fnmatch
from onegov.core.orm import Base, SessionManager
from onegov.core.security import Public
from onegov.core.utils import scan_morepath_modules
from onegov.server.config import Config
from onegov.server.core import Server
from sqlalchemy.pool import NullPool
from uuid import uuid4
from webtest import TestApp as Client


#: :class:`GroupContext` settings which may be overriden by commands
CONTEXT_SPECIFIC_SETTINGS = (
    'default_selector',
    'creates_path',
    'singular',
    'matches_required'
)


class GroupContextGuard(object):
    """ Contains methods which abort the commandline program if any condition
    is not met.

    Used as a mixin in :class:`GroupContext`.

    """
    def validate_guard_conditions(self, click_context):
        matches = tuple(self.matches)

        for name, method in inspect.getmembers(self, inspect.ismethod):
            if name.startswith('abort'):
                method(click_context, matches)

    def abort_if_no_selector(self, click_context, matches):
        if not self.selector:
            click.secho("Available selectors:")

            for selector in self.available_selectors:
                click.secho(" - {}".format(selector))

            abort("No selector provided, aborting.")

    def abort_if_no_subcommand(self, click_context, matches):
        if click_context.invoked_subcommand is None:
            click.secho("Paths matching the selector:")

            for match in matches:
                click.secho(" - {}".format(match))

            abort("No subcommand provided, aborting.")

    def abort_if_no_match(self, click_context, matches):
        if self.matches_required and not matches:
            click.secho("Available selectors:")

            for selector in self.available_selectors:
                click.secho(" - {}".format(selector))

            abort("Selector doesn't match any paths, aborting.")

    def abort_if_not_singular(self, click_context, matches):
        if self.singular and len(matches) > 1:
            click.secho("Paths matching the selector:")

            for match in matches:
                click.secho(" - {}".format(match))

            abort("The selector must match a single path, aborting.")

    def abort_if_no_create_path(self, click_context, matches):
        if self.creates_path:
            if len(matches) > 1:
                abort("This selector may not reference an existing path")

            if len(self.selector.lstrip('/').split('/')) != 2:
                abort("This selector must reference a full path")

            if '*' in self.selector:
                abort("This selector may not contain a wildcard")


class GroupContext(GroupContextGuard):
    """ Provides access to application configs for group commands.

    :param selector:
        Selects the applications which should be captured by a
        :func:`command_group`.

        See :ref:`core-commands` for more documentation about
        selectors.

    :param config:
        The targeted onegov.yml file or an equivalent dictionary.

    :param default_selector:
        The selector used if none is provided. If not given, a selector
        *has* to be provided.

    :param creates_path:
        True if the given selector doesn't exist yet, but will be created.
        Commands which use this setting are expected to take a single path
        (no wildcards) and to create it during their runtime.

        Implies `singular` and `matches_required`.

    :param singular:
        True if the selector may not match multiple applications.

    :param matches_required:
        True if the selector *must* match at least one application.

    """

    def __init__(self, selector, config, default_selector=None,
                 creates_path=False, singular=False, matches_required=True):

        if isinstance(config, dict):
            self.config = Config(config)
        else:
            self.config = Config.from_yaml_file(config)

        self.selector = selector or default_selector
        self.creates_path = creates_path

        if self.creates_path:
            self.singular = True
            self.matches_required = False
        else:
            self.singular = singular
            self.matches_required = matches_required

    def unbound_session_manager(self, dsn):
        """ Returns a session manager *not yet bound to a schema!*. """

        return SessionManager(dsn=dsn, base=Base, pool_config={
            'poolclass': NullPool
        })

    def available_schemas(self, appcfg):
        """ Returns all available schemas, if the application is database
        bound.

        """

        if 'dsn' not in appcfg.configuration:
            return []

        mgr = self.unbound_session_manager(appcfg.configuration['dsn'])
        return mgr.list_schemas(limit_to_namespace=appcfg.namespace)

    def match_to_path(self, match):
        """ Takes the given match and returns the application path used in
        http requests.

        """
        match = match.lstrip('/')

        if '/' in match:
            namespace, id = match.split('/')[:2]
        else:
            namespace, id = match, ''

        for appcfg in self.config.applications:
            if appcfg.namespace == namespace:
                return appcfg.path.replace('*', id).rstrip('/')

    @property
    def appcfgs(self):
        """ Returns the matching appconfigs.

        Since there's only one appconfig per namespace, we ignore the path
        part of the selector and only focus on the namespace::

            /namespace/application_id

        """
        namespace_selector = self.selector.lstrip('/').split('/')[0]

        for appcfg in self.config.applications:
            if fnmatch(appcfg.namespace, namespace_selector):
                yield appcfg

    @property
    def available_selectors(self):
        """ Generates a list of available selectors.

        The list doesn't technically exhaust all options, but it returns
        all selectors targeting a single application as well as all selectors
        targeting a namespace by wildcard.

        """
        selectors = list(self.all_wildcard_selectors)
        selectors.extend(self.all_specific_selectors)

        return sorted(selectors)

    @property
    def all_wildcard_selectors(self):
        """ Returns all selectors targeting a namespace by wildcard. """

        for appcfg in self.config.applications:
            if appcfg.path.endswith('*'):
                yield '/' + appcfg.namespace + '/*'

    @property
    def all_specific_selectors(self):
        """ Returns all selectors targeting an application directly. """

        for appcfg in self.config.applications:
            if not appcfg.path.endswith('*'):
                yield '/' + appcfg.namespace
            else:
                for schema in self.available_schemas(appcfg):
                    yield '/' + schema.replace('-', '/')

    @property
    def matches(self):
        """ Returns the specific selectors matching the context selector.

        That is, a combination of namespace / application id is returned.
        Since we only know an exhaustive list of application id's *if* we have
        a database connection this is currently limited to applications with
        one. Since we do not have any others yet that's fine.

        However if we implement a database-less application in the future which
        takes wildcard ids, we need some way to enumerate those ids.

        See https://github.com/OneGov/onegov.core/issues/13
        """
        if self.selector:
            for selector in self.all_specific_selectors:
                if fnmatch(selector, self.selector):
                    yield selector

        if self.creates_path:
            yield self.selector.rstrip('/')


def get_context_specific_settings(context):
    """ Takes the given *click* context and extracts all context specific
    settings from it.

    """

    if not context.invoked_subcommand:
        return {}

    # The context settings are stored on the command though they are actually
    # click's settings, not ours. Upon inspection we need to transfer them
    # here as a result. It's basically a piggy back ride.
    subcommand = context.command.commands[context.invoked_subcommand]
    if not hasattr(subcommand, 'onegov_settings'):
        subcommand.onegov_settings = {
            key: subcommand.context_settings.pop(key)
            for key in CONTEXT_SPECIFIC_SETTINGS
            if key in subcommand.context_settings
        }

    return subcommand.onegov_settings


#: Decorator to acquire the group context on a command::
#:
#:     @cli.command()
#:     @pass_group_context()
#:     def my_command(group_context):
#:         pass
#:
pass_group_context = click.make_pass_decorator(GroupContext, ensure=True)


def command_group():
    """ Generates a click command group for individual modules.

    Each individual module may have its own command group from which to run
    commands to. Read `<http://click.pocoo.org/6/commands/>`_ to learn more
    about command groups.

    The returned command group will provide the individual commands with
    an optional list of applications to operate on and it allows commands
    to return a callback function which will be invoked with the app config
    (if available), an application instance and a request.

    That is to say, the command group automates setting up a proper request
    context.

    """

    @click.group(invoke_without_command=True)
    @click.option(
        '--select', default=None,
        help="Selects the applications this command should be applied to")
    @click.option(
        '--config', default='onegov.yml',
        help="The onegov config file")
    def command_group(select, config):
        try:
            context = click.get_current_context()
            context_settings = get_context_specific_settings(context)
            context.obj = GroupContext(select, config, **context_settings)
            context.obj.validate_guard_conditions(context)
        except sqlalchemy.exc.OperationalError as e:
            if "Connection refused" in ' '.join(e.args):
                print(e)
                sys.exit(1)
            else:
                raise

    @command_group.resultcallback()
    def process_results(processor, select, config):
        """ Calls the function returned by the command once for each
        application matching the selector.

        Uses a proper request/application context for ease of use.

        """

        if not processor:
            return

        group_context = click.get_current_context().obj

        # load all applications into the server
        view_path = uuid4().hex
        applications = []

        for appcfg in group_context.appcfgs:

            class CliApplication(appcfg.application_class):

                def is_allowed_application_id(self, application_id):

                    if group_context.creates_path:
                        return True

                    return super().is_allowed_application_id(application_id)

                def configure_debug(self, **cfg):
                    # disable debug options in cli (like query output)
                    pass

            @CliApplication.path(path=view_path)
            class Model(object):
                pass

            @CliApplication.view(model=Model, permission=Public)
            def run_command(self, request):
                processor(request, request.app)

            @CliApplication.setting(section='cronjobs', name='enabled')
            def get_cronjobs_enabled():
                return False

            @CliApplication.setting_section(section='roles')
            def get_roles_setting():
                # override the security settings -> we need the public
                # role to work for anonymous users, even if the base
                # application disables that
                return {
                    'anonymous': {Public},
                }

            scan_morepath_modules(CliApplication)
            CliApplication.commit()

            applications.append({
                'path': appcfg.path,
                'application': CliApplication,
                'namespace': appcfg.namespace,
                'configuration': appcfg.configuration
            })

        server = Server(
            Config({
                'applications': applications,
                'logging': group_context.config.logging
            }),
            configure_morepath=False
        )

        # call the matching applications
        client = Client(server)

        matches = list(group_context.matches)

        for match in matches:
            path = group_context.match_to_path(match)
            client.get(path + '/' + view_path)

    return command_group


def abort(msg):
    """ Prints the given error message and aborts the program with a return
    code of 1.

    """
    click.secho(msg, fg='red')

    sys.exit(1)



""" Generic associations are an interesting SQLAlchemy design pattern
that allow for collections of records to be attached to other models by
mere inheritance.

Generic associations are useful in situations where a model should be
attachable to a number of third-party models. For example, we might say that
a comment model should be attacheable to any other model (say a page or a
ticket). In such an instance generic associations offer the benefit of
being simple to integrate on the third-party model.

Additionally we can define generic methods that work on all third-party
models that inherit from the associated model (in our example imagine a
"Commentable" class that leads to the automatic attachment of comments to
Page/Ticket models).

See also:
https://github.com/zzzeek/sqlalchemy/blob/master/examples/
generic_associations/table_per_association.py

Note that you do not want to use these associations in lieue of simple
relationships between two models. The standard way of doing this leads to a
strong relationship on the database and is easier to change and reason about.

Generic associations are meant for generic usecases. These currently include
payments (any model may be payable) and files (any model may have files
attached). Other generic associations should be introduced along these
lines.

A single model may be associated to any number of other models. For example::

                            ┌─────────────┐
                      ┌─────│ Reservation │
    ┌────────────┐    │     └─────────────┘
    │  Payment   │◀───┤
    └────────────┘    │     ┌─────────────┐
                      └─────│    Form     │
                            └─────────────┘

Here, ``Payment`` is associable (through the ``Payable`` mixin).
``Reservation`` and ``Form`` in turn inherit from ``Payable``.

This all is probably best understood in an example:

        class Payment(Base, Associable):
            __tablename__ == 'payments'

        class Payable(object):
            payment = associated(Address, 'address', 'one-to-one')

        class Product(Base, Payable):
            __tablename__ == 'products'

This results in a product model which has a payment attribute attached to it.
The payments are stored in the ``payments`` table, products in the ``products``
table. The link between the two is established in the automatically created
``payments_for_products`` table.

"""

from collections import namedtuple
from onegov.core.orm.utils import QueryChain
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Table
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import joinedload
from sqlalchemy.orm import object_session
from sqlalchemy.orm import relationship


class RegisteredLink(namedtuple("RegisteredLink", (
    'cls',
    'table',
    'key',
    'attribute',
    'cardinality',
))):

    @property
    def class_attribute(self):
        return getattr(self.cls, self.attribute)


def associated(associated_cls, attribute_name, cardinality='one-to-many',
               uselist='auto'):
    """ Creates an associated attribute. This attribute is supposed to be
    defined on the mixin class that will establish the generic association
    if inherited by a model.

    :param associated_cls:
        The class which the model will be associated with.

    :param attribute_name:
        The name of the attribute used on the mixin class.

    :param cardinality:
        May be 'one-to-one', 'one-to-many' or 'many-to-many'. Cascades are
        used automatically, unless 'many-to-many' is used, in which case
        cascades are disabled.

    :param uselist:
        True if the attribute on the inheriting model is a list. Use 'auto'
        if this should be automatically decided depending on the cardinality.

    Example::

        class Adress(Base, Associable):
            pass

        class Addressable(object):
            address = associated(Address, 'address', 'one-to-one')

        class Company(Base, Addressable):
            pass

    """

    assert cardinality in ('one-to-one', 'one-to-many', 'many-to-many')

    if cardinality in ('one-to-one', 'one-to-many'):
        cascade = 'all, delete-orphan'
        single_parent = True
        passive_deletes = False
    else:
        cascade = False
        single_parent = False
        passive_deletes = True

    if uselist == 'auto':
        uselist = not cardinality.endswith('to-one')

    def descriptor(cls):
        name = '_for_'.join((associated_cls.__tablename__, cls.__tablename__))
        key = '{}_id'.format(cls.__tablename__)
        target = '{}.id'.format(cls.__tablename__)
        backref = 'linked_{}'.format(cls.__tablename__)

        association_key = associated_cls.__name__.lower() + '_id'
        association_id = associated_cls.id

        association = Table(
            name, cls.metadata,
            Column(key, ForeignKey(target), nullable=False),
            Column(association_key, ForeignKey(association_id), nullable=False)
        )

        assert issubclass(associated_cls, Associable)

        associated_cls.register_link(
            backref, cls, association, key, attribute_name, cardinality)

        return relationship(
            argument=associated_cls,
            secondary=association,
            backref=backref,
            single_parent=single_parent,
            cascade=cascade,
            uselist=uselist,
            passive_deletes=passive_deletes
        )

    return declared_attr(descriptor)


class Associable(object):
    """ Mixin to enable associations on a model. Only models which are
    associable may be targeted by :func:`associated`_

    """

    registered_links = None

    @classmethod
    def association_base(cls):
        """ Returns the model which directly inherits from Associable. """

        for parent in cls.__bases__:
            if parent is Associable:
                return cls
            if issubclass(parent, Associable):
                return parent.association_base()

        return cls

    @classmethod
    def register_link(cls, link_name, linked_class, table, key,
                      attribute, cardinality):
        """ All associated classes are registered through this method. This
        yields the following benefits:

        1. We gain the ability to query all the linked records in one query.
           This is hard otherwise as each ``Payable`` class leads to its own
           association table which needs to be queried separately.

        2. We are able to reset all created backreferences. This is necessary
           during tests. SQLAlchemy keeps these references around and won't
           let us re-register the same model multiple times (which outside
           of tests is completely reasonable).

        """
        base = cls.association_base()

        if not base.registered_links:
            base.registered_links = {}

        base.registered_links[link_name] = RegisteredLink(
            cls=linked_class,
            table=table,
            key=key,
            attribute=attribute,
            cardinality=cardinality
        )

    @property
    def links(self):
        """ Returns a query chain with all records of all models which attach
        to the associable model.

        """

        session = object_session(self)

        def query(link):
            column = getattr(link.cls, link.attribute)

            q = session.query(link.cls)

            if column.property.uselist:
                q = q.filter(column.contains(self))
            else:
                q = q.filter(column == self)

            return q.options(joinedload(column))

        return QueryChain(tuple(
            query(link) for link in self.registered_links.values()
        ))



import time

from onegov.core.errors import AlreadyLockedError
from onegov.core.locking import lock
from onegov.core.orm import SessionManager
from sqlalchemy.ext.declarative import declarative_base
from threading import Thread


class LockingThread(Thread):

    def __init__(self, mgr, namespace, key, schema):
        Thread.__init__(self)
        self.mgr = mgr
        self.namespace = namespace
        self.key = key
        self.schema = schema

    def run(self):
        self.mgr.set_current_schema(self.schema)

        try:
            with lock(self.mgr.session(), self.namespace, self.key):
                time.sleep(0.1)
        except Exception as e:
            self.exception = e
        else:
            self.exception = None


def test_locking_smoke(session):
    with lock(session, 'foo', 'bar'):
        pass


def test_threaded_locking(postgres_dsn):
    Base = declarative_base()

    session_manager = SessionManager(postgres_dsn, Base)
    session_manager.set_current_schema('foo')

    threads = [
        LockingThread(session_manager, 'namespace', 'key', 'foo'),
        LockingThread(session_manager, 'namespace', 'key', 'foo')
    ]

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    exceptions = [t.exception for t in threads if t.exception]

    assert len(exceptions) == 1
    assert isinstance(exceptions[0], AlreadyLockedError)


def test_multischema_threaded_locking(postgres_dsn):
    Base = declarative_base()

    session_manager = SessionManager(postgres_dsn, Base)
    session_manager.set_current_schema('foo')

    threads = [
        LockingThread(session_manager, 'namespace', 'key', 'foo'),
        LockingThread(session_manager, 'namespace', 'key', 'bar')
    ]

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    exceptions = [t.exception for t in threads if t.exception]

    assert len(exceptions) == 1
    assert isinstance(exceptions[0], AlreadyLockedError)



import mailthon.headers as headers
import os.path
import re

from email.mime.multipart import MIMEMultipart
from html2text import html2text
from mailbox import Maildir, MaildirMessage
from mailthon.enclosure import HTML, PlainText, Attachment
from mailthon.envelope import Envelope as BaseEnvelope
from mailthon.postman import Postman


VALID_PLAINTEXT_CHARACTERS = re.compile(r"""
    [
        \d  # decimals
        \w  # words
        \n  # new lines

        # emojis
        \U00002600-\U000027BF
        \U0001f300-\U0001f64F
        \U0001f680-\U0001f6FF
    ]+
""", re.VERBOSE)


def convert_to_plaintext(html):

    # filter out duplicated lines and lines without any text
    lines = html2text(html, '', bodywidth=0).splitlines()
    lines = (l.strip() for l in lines)
    lines = (l for l in lines if VALID_PLAINTEXT_CHARACTERS.search(l))

    # use double newlines to get paragraphs
    plaintext = '\n\n'.join(lines)

    # in an attempt to create proper markdown html2text will escape
    # dots. Since markdown is not something we care about here, we undo that
    plaintext = plaintext.replace('\\.', '.')

    return plaintext


def email(sender=None, receivers=(), cc=(), bcc=(),
          subject=None, content=None, encoding='utf8',
          attachments=()):
    """
    Creates an Envelope object with a HTML *content*, as well as a *plaintext*
    alternative generated from the HTML content.

    :param content: HTML content.
    :param encoding: Encoding of the email.
    :param attachments: List of filenames to
        attach to the email.

    Note: this is basically a copy of :func:`mailthon.api.email`, though it
    adds the plaintext alternative.
    """

    if content:
        # turn the html email into a plaintext representation
        # this leads to a lower spam rating
        plaintext = convert_to_plaintext(content)
    else:
        plaintext = None

    # According to RFC 2046, the last part of a multipart message, in this case
    # the HTML message, is best and preferred
    enclosure = [
        PlainText(plaintext, encoding),
        HTML(content, encoding),
    ]
    enclosure.extend(Attachment(k) for k in attachments)
    return Envelope(
        headers=[
            headers.subject(subject),
            headers.sender(sender),
            headers.to(*receivers),
            headers.cc(*cc),
            headers.bcc(*bcc),
            headers.date(),
            headers.message_id(),
        ],
        enclosure=enclosure,
    )


class Envelope(BaseEnvelope):
    """ Changes the mailthon envelope to use mime-multipart/alternative.

    """

    def mime(self):
        mime = MIMEMultipart('alternative')
        for item in self.enclosure:
            mime.attach(item.mime())
        self.headers.prepare(mime)
        return mime


class MaildirTransport(object):
    """ A transport that pretends to be like python's smtplib.SMTP transport,
    but actually just stores files into a maildir.

    """

    def __init__(self, host, port, maildir):
        # host and port exist for mailthon compatibility, they are ignored
        self.maildir = Maildir(maildir, create=True)

        for d in (os.path.join(maildir, d) for d in ('new', 'cur', 'tmp')):
            if not os.path.exists(d):
                os.makedirs(d)

    def sendmail(self, from_addr, to_addrs, message):
        self.maildir.add(MaildirMessage(message))

        return {}

    def noop(self):
        return 250, ""

    def ehlo(self):
        pass

    def quit(self):
        pass


class MaildirPostman(Postman):
    """ A mailthon postman that stores mails to a maildir. """
    transport = MaildirTransport
